<?php
namespace Controller;

use Util\InstanceProvider;
use Util\Utilities;
use Zeedhi\DevOps\Util\Utilities AS DevOpsUtilities ;
use Zeedhi\Framework\DataSource\Manager;
use Zeedhi\Framework\DTO;
use Zeedhi\Framework\DTO\Response;
use Zeedhi\Framework\DTO\Request;
use Zeedhi\Framework\DataSource\DataSet;


class Classification extends \Generated\Controller\Classification
{

    /**
     * @var Utilities
     */
    private $utilities;
    /**
     * @var \Service\Classification
     */
    private $classificationService;

    public function __construct(Manager $dataSourceManager, Utilities $utilities,
                                \Service\Classification $classificationService)
    {
        $this->utilities = $utilities;
        parent::__construct($dataSourceManager);
        $this->classificationService = $classificationService;
    }

    public function find(DTO\Request\Filter $request, DTO\Response $response)
    {
        try {
            $request->getFilterCriteria()->addCondition("IS_TASK", "IN", array("Y"));
            $conditions = $request->getFilterCriteria()->getConditions();
            $dataSet = $this->classificationService->findClassifications($conditions);
            $response->addDataset(new DataSet('classification', $dataSet));
        } catch (\Exception $e) {
            throw new \Exception($e->getMessage() . ' Code:' . $e->getCode());
        }
    }

    public function findTaskClassification(DTO\Request\Filter $request, DTO\Response $response)
    {
        try {
            $request->getFilterCriteria()->addCondition("IS_TASK", "IN", array("Y"));
            $conditions = $request->getFilterCriteria()->getConditions();
            $dataSet = $this->classificationService->findClassifications($conditions);
            $response->addDataset(new DataSet('taskClassification', $dataSet));
        } catch (\Exception $e) {
            throw new \Exception($e->getMessage() . ' Code:' . $e->getCode());
        }
    }

    public function delete(DTO\Request\DataSet $request, DTO\Response $response)
    {
        try {
            $row = $request->getDataSet()->getRows()[0];
            $this->classificationService->deleteClassification($row['ID']);
        } catch (\Exception $e) {
            throw new \Exception($e->getMessage() . ' Code:' . $e->getCode());
        }
    }

    public function findClassificationForTask(DTO\Request\Filter $request, DTO\Response $response)
    {
        try {
            $conditions = $request->getFilterCriteria()->getConditions();
            $projectStatus = Utilities::getValueInFilterDataSourceByName($conditions,'PROJECT_STATUS');
            $projectType = Utilities::getValueInFilterDataSourceByName($conditions,'PROJECT_TYPE');
            $filterString = Utilities::getFilterTableColumn($conditions, 'LABEL');
            $dataSet = $this->classificationService->findClassificationForTask($projectStatus,$projectType, $filterString);
            $response->addDataset(new DataSet('classification', $dataSet));
        } catch (\Exception $e) {
            throw new \Exception($e->getMessage() . ' Code:' . $e->getCode());
        }
    }

    public function findClassificationForMeeting(DTO\Request\Filter $request, DTO\Response $response)
    {
        try {
            $conditions = $request->getFilterCriteria()->getConditions();
            $project = Utilities::getValueInFilterDataSourceByName($conditions,'PROJECT',true);
            $filterString = Utilities::getFilterTableColumn($conditions, 'LABEL');
            $dataSet = InstanceProvider::getService()->getClassification()->findClassificationForMeeting($project,$filterString);
            $response->addDataset(new DataSet('classification', $dataSet));
        } catch (\Exception $e) {
            throw new \Exception($e->getMessage() . ' Code:' . $e->getCode());
        }
    }

    public function save(DTO\Request\DataSet $request, DTO\Response $response)
    {
        try {
            $row = $request->getDataSet()->getRows()[0];
            $isProject = $row["IS_PROJECT"];
            $isTask = $row["IS_TASK"];
            $id = $row["ID"];
            $label = $row["LABEL"];
            $allowedProjectType = $row["ALLOWED_PROJECT_TYPE"];
            $allowedProjectStatus = $row["ALLOWED_PROJECT_STATUS"];
            $notAllowedProjectType = $row["NOT_ALLOWED_PROJECT_TYPE"];
            $notAllowedProjectStatus = $row["NOT_ALLOWED_PROJECT_STATUS"];
            $estimatedTimeProportion = $row["ESTIMATED_TIME_PROPORTION"];
            $usedInTimeControl = $row["USED_IN_TIME_CONTROL"];
            $usedInNews = $row["USED_IN_NEWS"];
            $control = empty($row["CONTROL"]) ? null : $row["CONTROL"];
            $divideDuration = empty($row["DIVIDE_DURATION"]) ? null : $row["DIVIDE_DURATION"];
            $accessProfile = empty($row["NRPERFILACESSO"]) ? array() : $row["NRPERFILACESSO"];

            $response->addDataset(new DataSet('classification', $this->classificationService->saveClassification($id, $label, $isProject, $isTask, $control,
                $allowedProjectType, $allowedProjectStatus, $notAllowedProjectType, $notAllowedProjectStatus, $estimatedTimeProportion, $divideDuration, $usedInTimeControl, $accessProfile, $usedInNews)));
        } catch (\Exception $e) {
            throw new \Exception($e->getMessage() . ' Code:' . $e->getCode());
        }
    }
    
}

