<?php
/**
 * Created by PhpStorm.
 * User: tuliomagalhaes
 * Date: 01/08/2016
 * Time: 15:51
 */

namespace Service;

use Doctrine\ORM\EntityManager;
use Zeedhi\DevOps\Model\Projectmanagement\Repositories;
use Zeedhi\DevOps\Model\Projectmanagement\Entities;
use Zeedhi\DevOps\Util\Utilities;

class ClassificationProjectType {

    /** @var Repositories\ClassificationProjectType */
    protected $classificationProjectTypeRepository;

    /** @var Repositories\ProjectType */
    protected $projectTypeRepository;

    /** @var EntityManager*/
    private $entityManager;

    /**@var Utilities */
    private $environment;

    /** @var \Factory\Classification*/
    private $classificationProjectTypeFactory;

    public function __construct(EntityManager $entityManager, \Util\Utilities $environment,
                                \Factory\ClassificationProjectType $classificationProjectTypeFactory)
    {
        $this->entityManager = $entityManager;
        $this->environment = $environment;
        $this->classificationProjectTypeFactory = $classificationProjectTypeFactory;
        $this->classificationProjectTypeRepository = $this->entityManager->getRepository(Utilities::PROJECT_MANAGEMENT_ENTITY_DIR . 'ClassificationProjectType');
        $this->projectTypeRepository = $this->entityManager->getRepository(Utilities::PROJECT_MANAGEMENT_ENTITY_DIR . 'ProjectType');
    }

    public function saveArrayProjectType($classification, array $arrayProjectType, $type)
    {

        foreach ($arrayProjectType as $projectType) {
            if (!empty($projectType)) {
                $classificationProjectType = $this->classificationProjectTypeRepository->findOneBy(
                    array('projectType' => $projectType,'classification' => $classification, 'type' => $type));
                if (empty($classificationProjectType)) {
                    $this->createClassificationProjectType($classification, $projectType, $type);
                }
            }
        }

        $this->entityManager->flush();
        /** @var Entities\ClassificationProjectType[] $arrayClassificationProjectType */
        $arrayClassificationProjectType = $this->classificationProjectTypeRepository->findBy(
            array('classification' => $classification, 'type' => $type));
        $arrayProjectTypeId = Utilities::getArrayByArrayObjectWithGet($arrayProjectType);
        foreach ($arrayClassificationProjectType as $classificationProjectType) {
            if (!in_array($classificationProjectType->getProjectType()->getId(), $arrayProjectTypeId)) {
                $this->entityManager->remove($classificationProjectType);
            }
        }
        $this->entityManager->flush();
    }

    public function saveArrayAllowedProjectTypeId($classification, array $projectTypeIds)
    {
        $taskStatus = array();
        foreach ($projectTypeIds as $projectTypeId) {
            $taskStatus[] = $this->projectTypeRepository->find($projectTypeId);
        }
        $this->saveArrayProjectType($classification, $taskStatus, Entities\ClassificationProjectType::ALLOWED_TYPE);
    }

    public function saveArrayNotAllowedProjectTypeId($classification, array $projectTypeIds)
    {
        $projectType= array();
        foreach ($projectTypeIds as $projectTypeId) {
            $projectType[] = $this->projectTypeRepository->find($projectTypeId);
        }
        $this->saveArrayProjectType($classification, $projectType, Entities\ClassificationProjectType::NOT_ALLOWED_TYPE);
    }

    public function createClassificationProjectType(Entities\Classification $classification, Entities\ProjectType $projectType, $type)
    {
        $classificationProjectType = $this->classificationProjectTypeFactory->createClassificationProjectType($classification,
            $projectType, $type);
        $this->entityManager->persist($classificationProjectType);
    }

    public function deleteClassificationProjectType(Entities\Classification $classification)
    {
        $classificationProjectTypes = $this->classificationProjectTypeRepository->findBy(array('classification' => $classification));
        foreach ($classificationProjectTypes as $classificationProjectType) {
            $this->entityManager->remove($classificationProjectType);
        }
        $this->entityManager->flush();
    }
} 