<?php

namespace Service;

use Doctrine\ORM\EntityManager;
use Service\Calendar\Restriction\Strategy\RestrictionFactory;
use Service\Calendar\TaskCalendar;
use Service\Queue\Task\Consumer;
use Util\Messages;
use Util\Utilities;
use Zeedhi\DevOps\Util\Utilities as DevOpsUtilities;
use Zeedhi\DevOps\Model\Configurationmanager\Entities\Product;
use Zeedhi\DevOps\Model\Configurationmanager\Entities\Release;
use Zeedhi\DevOps\Model\Configurationmanager\Entities\Package;
use Zeedhi\DevOps\Model\Configurationmanager\Entities\PackageContents;
use Zeedhi\DevOps\Model\Configurationmanager\Entities\Subrelease;
use Zeedhi\DevOps\Model\Util\LogicalDeleted;
use Zeedhi\Framework\Controller\Exception;
use Zeedhi\Framework\DTO\Response\Message;
use Zeedhi\Framework\HTTP\Kernel;
use Zeedhi\DevOps\Model\Projectmanagement\Entities;
use Zeedhi\DevOps\Model\Projectmanagement\Repositories;
use Zeedhi\DevOps\Model\Profile\Entities\ZhUser;
use Util\InstanceProvider;
use Zeedhi\DevOps\Model\Devops\Entities as DevopsEntities;
use Zeedhi\DevOps\Model\Configurationmanager\Entities as ConfigurationmanagerEntities;

/**
 * Class Project
 *
 * @package Service
 */
class Project
{
    protected $entityManager;
    protected $environment;

    const ACTIVE = 'A';
    const COMPANY_TYPE = 'C';

    const IS_BOARD = 'Y';
    /**
     * types in projectResouce
     */
    const WORK_TYPE = 'W';
    const SPONSOR_TYPE = 'S';
    const PROJECT_MANAGER_TYPE = 'M';

    const IS_PARENT_TRUE = 'Y';
    const IS_PARENT_FALSE = 'N';

    const MNEMONIC_SEPARATOR = '[';

    /** all CONTROL of PROJECT TYPE */
    const PROJECT_TYPE_ISSUE = Entities\Project::PROJECT_TYPE_ISSUE;
    const PROJECT_TYPE_PRODUCT = Entities\Project::PROJECT_TYPE_PRODUCT;
    const PROJECT_TYPE_STANDARD = Entities\Project::PROJECT_TYPE_STANDARD;
    const PROJECT_TYPE_RELEASE = Entities\Project::PROJECT_TYPE_RELEASE;
    const PROJECT_TYPE_SUB_RELEASE = Entities\Project::PROJECT_TYPE_SUB_RELEASE;
    const PROJECT_TYPE_ERROR = Entities\Project::PROJECT_TYPE_ERROR;
    const PROJECT_TYPE_MEETING = Entities\Project::PROJECT_TYPE_MEETING;
    const PROJECT_TYPE_MEETING_OCCURRENCES = Entities\Project::PROJECT_TYPE_MEETING_OCCURRENCES;
    const PROJECT_TYPE_MODEL_RESTRICTED = Entities\Project::PROJECT_TYPE_MODEL_RESTRICTED;
    const PROJECT_TYPE_MODEL_FREE = Entities\Project::PROJECT_TYPE_MODEL_FREE;
    const PROJECT_TYPE_TEST = Entities\Project::PROJECT_TYPE_TEST;

    /** IS_FIX_RELEASE */
    const IS_FIX_RELEASE = 'Y';
    const NOT_IS_FIX_RELEASE = 'N';

    /** Type of DevOps Associations */
    const TYPE_PRODUCT = 'P';
    const TYPE_RELEASE = 'R';
    const TYPE_SUBRELEASE = 'S';

    private static $DATA_SET_EXCEPTION_LEAF = array(
        "DI" => "Deve Iniciar em",
        "DT" => "Deve Terminar em",
        "IA" => "Não Iniciar Antes de",
        "ID" => "Não Iniciar Depois de",
        "TA" => "Não Terminar Antes de",
        "TD" => "Não Terminar Depois de",
        "MB" => "O Mais Breve Possível",
        "MT" => "O Mais Tarde Possível");

    private static $DATA_SET_GENERATED_TYPE_RELATIONSHIP_TASK = array(
        Entities\TaskRelationship::CONFLICT_TYPE,
        Entities\TaskRelationship::ERROR_TYPE,
        Entities\TaskRelationship::ERROR_MR_TYPE,
        Entities\TaskRelationship::TEST_TYPE,
        Entities\TaskRelationship::MERGE_REQUEST_TYPE,
        Entities\TaskRelationship::REPLICATION_TYPE,
        Entities\TaskRelationship::ORIGIN_TYPE
    );
    private static $PROJECT_TYPE_COPY_IN_TIMELINE = array(
        Entities\ProjectType::TYPE_PROJECT_CONTROL_RELEASE,
        Entities\ProjectType::TYPE_PROJECT_CONTROL_SUB_RELEASE,
        Entities\ProjectType::TYPE_PROJECT_CONTROL_ERROR,
    );

    const PROJECT_CONCLUDED_STATUS = Entities\Project::PROJECT_CONCLUDED_STATUS;
    const PROJECT_PAUSED_STATUS = Entities\Project::PROJECT_PAUSED_STATUS;
    const PROJECT_IN_PROGRESS_STATUS = Entities\Project::PROJECT_IN_PROGRESS_STATUS;
    const PROJECT_NOT_INITIALIZED_STATUS = Entities\Project::PROJECT_NOT_INITIALIZED_STATUS;
    const PROJECT_CANCELED_STATUS = Entities\Project::PROJECT_CANCELED_STATUS;
    public $projectTypeIds;

    public function __construct($projectTypeIds)
    {

        $this->environment = InstanceProvider::getEnvironment();
        $this->entityManager = InstanceProvider::getEntityManager();
        $this->projectTypeIds = $projectTypeIds;

        //$this->taskService = $taskService;
        //$this->taskPrecursor = $taskPrecursor;
        //$this->resourceProjectService = $resourceProjectService;
        //$this->allocationService = $allocationService;
        //$this->calendarService = $calendarService;
        //$this->projectFactory = $projectFactory;
        //$this->releaseFactory = $releaseFactory;
        //$this->productFactory = $productFactory;
        //$this->subReleaseFactory = $subReleaseFactory;
        //$this->resourceProjectFactory = $resourceProjectFactory;
        //$this->meetingFactory = $meetingFactory;

        //$this->repositoryService = $repositoryService;
        //$this->taskCalendar = $taskCalendar;
        //$this->attachmentService = $attachmentService;
        //$this->calendarDetailService = $calendarDetailService;
        //$this->consumer = $consumer;
        //$this->resourceService = $resourceService;
        //$this->projectRelationshipService = $projectRelationshipService;
        //$this->zhUserAllocationService = $zhUserAllocationService;

        //$this->projectRepository = $entityManager->getRepository(Utilities::ENTITY_DIR . 'Project');
        //$this->taskRepository = $entityManager->getRepository(Utilities::ENTITY_DIR . 'Task');
        //$this->structureRepository = $entityManager->getRepository(Utilities::ENTITY_DIR . 'Structure');
        //$this->attachmentRepository = $entityManager->getRepository(Utilities::ENTITY_DIR . 'Attachment');
        //$this->projectTypeRepository = $entityManager->getRepository(Utilities::ENTITY_DIR . 'ProjectType');
        //$this->resourceRepository = $entityManager->getRepository(Utilities::ENTITY_DIR . 'Resource');
        //$this->taskStatusRepository = $entityManager->getRepository(Utilities::ENTITY_DIR . 'TaskStatus');
        //$this->taskRelationshipRepository = $entityManager->getRepository(Utilities::ENTITY_DIR . 'TaskRelationship');
        //$this->classificationRepository = $entityManager->getRepository(Utilities::ENTITY_DIR . 'Classification');
        //$this->repositoryRepository = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'Repository');
        //$this->directoryRepository = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'Directory');
        //$this->calendarDatesRepository = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'CalendarDates');
        //$this->projectApprovalRepository = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'ProjectApproval');
        //$this->permissionRepository = $entityManager->getRepository(Utilities::ENTITY_DIR . 'Permission');
        //$this->resourceProjectRepository = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'ResourceProject');
        //$this->meetingRepository = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'Meeting');
        //$this->meetingProjectRepository = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'MeetingProject');
        //$this->releaseRepository = $this->entityManager->getRepository(Utilities::CONFIGURATION_ENTITY_DIR . 'Release');
        //$this->productRepository = $this->entityManager->getRepository(Utilities::CONFIGURATION_ENTITY_DIR . 'Product');
        //$this->subReleaseRepository = $this->entityManager->getRepository(Utilities::CONFIGURATION_ENTITY_DIR . 'Subrelease');
        //$this->packageRepository = $this->entityManager->getRepository(Utilities::CONFIGURATION_ENTITY_DIR . 'Package');
        //$this->packageContentsRepository = $this->entityManager->getRepository(Utilities::CONFIGURATION_ENTITY_DIR . 'PackageContents');
        //$this->productServerRepository = $this->entityManager->getRepository(Utilities::SERVER_ENTITY_DIR . 'ProductServer');
        //$this->changedFilesRepository = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'ChangedFiles');
        \Util\InstanceProvider::getLog()->setLog('instanciando project service',false,2);
    }

    public function recalculateProject($projectId)
    {
        $message = array(
            'class' => 'Project',
            'id' => $projectId,
        );
        $message = json_encode($message);
        InstanceProvider::getService()->getQueueTaskConsumer()->consumeCalendar($message);
    }

    public function recalculateTasks($taskId)
    {
        $message = array(
            'class' => 'Task',
            'id' => $taskId,
        );
        $message = json_encode($message);
        InstanceProvider::getService()->getQueueTaskConsumer()->consumeCalendar($message);
    }

    public static function arrayStatusNotModify(){
        return array(self::PROJECT_CANCELED_STATUS,self::PROJECT_CONCLUDED_STATUS);
    }
    public static function arrayStatusChildrenReplicate(){
        return array(
            self::PROJECT_CANCELED_STATUS,
            self::PROJECT_CONCLUDED_STATUS,
            Entities\TaskStatus::TASK_STATUS_CONTROL_IN_HOMOLOGATION
        );
    }
    public static function arrayProjectTypeDevOps(){
        return array(
            Entities\ProjectType::TYPE_PROJECT_CONTROL_RELEASE,
            Entities\ProjectType::TYPE_PROJECT_CONTROL_SUB_RELEASE,
            Entities\ProjectType::TYPE_PROJECT_CONTROL_ERROR
        );
    }
    public function findMyProjectByUser($conditions)
    {
        $userId = $this->environment->getCurrentUserId();
        $projectsAll = InstanceProvider::getRepository()->getProject()->findMyProjectByUser($userId, $conditions);

        $projectsAllIDs = array();
        foreach ($projectsAll as $project) {
            array_push($projectsAllIDs, $project['ID']);
        }
        foreach ($projectsAll as &$project) {
            if (!in_array($project['PARENT'], $projectsAllIDs)) {
                $project['PARENT'] = null;
            }
            $project['TIME_PLANNED'] = $project['TIME_PLANNED'] ? $project['TIME_PLANNED'] : null;
            if ($project['DURATION'] && $project['USETIMEPLANNING'] == 'Y' ) {
                $project['DURATION'] = $project['DURATION'] / (60 * 60) . 'H';
            }
            $project['IN_HOMOLOGATED'] =  $project['PROJECT_STATUS_CONTROL'] == Entities\TaskStatus::TASK_STATUS_CONTROL_IN_HOMOLOGATION;

        }
        return $projectsAll;
    }

    public function findAllProjectByUser($projectId)
{
        $userId = $this->environment->getCurrentUserId();
        $projectsAll = InstanceProvider::getRepository()->getProject()->findAllProjectByUser($userId, $projectId, $this->projectTypeIds);
        return Utilities::fixParentsForTreeGrid($projectsAll);
    }

    public function findMyProjectByUserInMeeting()
    {
        $conditions = $this->setConditionsMyProjectDefault();
        $organizationId = $this->environment->getCurrentOrganizationId();
        $projectsAll = InstanceProvider::getRepository()->getProject()->findMyProjectByUserMeeting($organizationId, $conditions);
        $projectsAll =  Utilities::fixParentsForTreeGrid($projectsAll);
        return $projectsAll;
    }

    public function findMyProjectByUserWithMeeting($conditions)
    {
        $userId = $this->environment->getCurrentUserId();
        $projectsAll = InstanceProvider::getRepository()->getProject()->findMyProjectByUserWithMeeting($userId, $conditions);

        $projectsAllIDs = array();
        foreach ($projectsAll as $project) {
            array_push($projectsAllIDs, $project['ID']);
        }
        foreach ($projectsAll as &$project) {
            if (!in_array($project['PARENT'], $projectsAllIDs)) {
                $project['PARENT'] = null;
            }
            if ($project['DURATION'] && $project['USETIMEPLANNING'] == 'Y' ) {
                $project['DURATION'] = $project['DURATION'] / (60 * 60) . 'H';
            }
        }
        return $projectsAll;
    }

    public function findProjectProductAndMeeting($filters = null){
        $filterString = Utilities::createQueryFiltered($filters);
        $userId = $this->environment->getCurrentUserId();
        $organizationId = $this->environment->getCurrentOrganizationId();
        $products = InstanceProvider::getRepository()->getProject()->findProjectProductAndMeeting($userId, $filterString);
        $notModelProducts = DevOpsUtilities::removeModels($products,$organizationId);
        return $notModelProducts;
    }

    public function getAllTaskSuperAllocation(array $allAllocationsDay, $project, $end, $begin, $duration, $precursorId = 0)
    {
        $isSuperAllocation = array();
        $calendarService = InstanceProvider::getService()->getCalendar();
        $calendarDates = $calendarService->findCalendarDatesForCalculate($project, $begin, $end);
        foreach ($allAllocationsDay as $allAllocationDay) {
            if ($allAllocationDay['DAY'] == $end || $allAllocationDay['DAY'] == $begin) {
                if ($precursorId != 0) {
                    /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Task $precursorTask */
                    $precursorTask = $this->entityManager->find('\Zeedhi\DevOps\Model\Projectmanagement\Entities\Task', $precursorId);
                    $timePrecursor = $precursorTask->getTimeRemaining();
                } else {
                    $timePrecursor = 0;
                }
                $dayHour = $calendarService->getDayHourByDate($allAllocationDay['DAY'], $calendarDates);
                if($dayHour != 0){
                    $endDuration = ($timePrecursor + $duration) % $dayHour;
                    $beginDuration = ($duration - $endDuration) % $dayHour;
                    if ($allAllocationDay['DAY'] == $end) {
                        if ($endDuration == 0 || ($allAllocationDay['HOUR_TOTAL_DAY'] + $endDuration) > $dayHour) {
                            $isSuperAllocation = array_merge($isSuperAllocation, explode(', ', $allAllocationDay['NAME_TASKS']));
                        }
                    } else if ($beginDuration + $allAllocationDay['HOUR_TOTAL_DAY'] > $dayHour) {
                        $isSuperAllocation = array_merge($isSuperAllocation, explode(', ', $allAllocationDay['NAME_TASKS']));
                    }
                }
            } else {
                $isSuperAllocation = array_merge($isSuperAllocation, explode(', ', $allAllocationDay['NAME_TASKS']));
            }
        }
        return $isSuperAllocation;
    }

    public function getAllPrecursorIds($precursorId)
    {
        if (empty($precursorId)) {
            $precursorIds = array($precursorId);
        } else {
            $precursorIds = Utilities::getArrayFromAssociativeArray(InstanceProvider::getRepository()->getTaskRelationship()->getAllPrecursors($precursorId));
        }
        return $precursorIds;
    }

    public function validateMaxDurationDay($taskId, $resourceIds, $betweenDates, $duration, $projectId, $precursorId = 0)
    {

        $allAllocations = array();
        $allTasks = array();
        if(!empty($resourceIds)){
            $begin = $betweenDates[0];
            $end = $betweenDates[1];
            $betweenDatesFilter = "AND DAY BETWEEN TO_DATE( '$begin 00:00:00' ,'DD/MM/YYYY HH24:MI:SS')
                                AND TO_DATE( '$end 23:59:59' ,'DD/MM/YYYY HH24:MI:SS')";
            $projectRepository = InstanceProvider::getRepository()->getProject();
            /** @var Entities\Project $project */
            $project = $projectRepository->find($projectId);
            $projectId = $project->getId();
            $allPrecursor = $this->getAllPrecursorIds($precursorId);
            $resourceRepository = InstanceProvider::getRepository()->getResource();
            foreach ($resourceIds as $resourceId) {
                if ($resourceId) {
                    /** @var Entities\Resource $resource */
                    $resource = $resourceRepository->find($resourceId);
                    $allAllocationsDay = $projectRepository->findAllAllocationDayProject($projectId, $taskId,
                        $betweenDatesFilter, $resourceId, $allPrecursor);
                    $tasks = $this->getAllTaskSuperAllocation($allAllocationsDay, $project, $end, $begin, $duration, $precursorId);
                    if (!empty($tasks)) {
                        array_push($allAllocations, $resource->getName());
                        $allTasks = array_merge($allTasks, $tasks);
                    }
                }
            }
            $allTasks = array_unique($allTasks);
            sort($allTasks);
        }
        if (empty($allAllocations)) {
            $dataSet = array();
        } else {
            $dataSet = array('resource' => $allAllocations, 'task' => $allTasks);
        }
        return $dataSet;
    }

    public function findAssociateProjectByUser($conditions)
    {

        $userId = $this->environment->getCurrentUserId();
        $projectsAll = InstanceProvider::getRepository()->getProject()->findAssociateProjectByUser($userId, $conditions);
        $projectsAll =  Utilities::fixParentsForTreeGrid($projectsAll);
        foreach ($projectsAll as &$project) {

            if ($project['DURATION'] && $project['USETIMEPLANNING'] == 'Y' ) {
                $project['DURATION'] = $project['DURATION'] / (60 * 60) . 'H';
            }
        }
        return $projectsAll;
    }

    public function findProjectByUserPMA()
    {

        $userId = $this->environment->getCurrentUserId();
        $projectsAll = InstanceProvider::getRepository()->getProject()->findProjectByUserPMA($userId);
        $projectsAllIDs = array();
        foreach ($projectsAll as $project) {
            array_push($projectsAllIDs, $project['ID']);
        }
        foreach ($projectsAll as &$project) {
            if (!in_array($project['PARENT'], $projectsAllIDs)) {
                $project['PARENT'] = null;
            }
        }
        return $projectsAll;
    }

    public function findPublicProjectByUser()
    {

        $userId = $this->environment->getCurrentUserId();
        $projectsAll = InstanceProvider::getRepository()->getProject()->findPublicProjectByUser($userId);
        $projectsAllIDs = array();
        foreach ($projectsAll as $project) {
            array_push($projectsAllIDs, $project['ID']);
        }
        foreach ($projectsAll as &$project) {
            if (!in_array($project['PARENT'], $projectsAllIDs)) {
                $project['PARENT'] = null;
            }
        }
        return $projectsAll;
    }

    public function findAllProjectParent($projectId)
    {
        $projectsAll = InstanceProvider::getRepository()->getProject()->getAllProjectParent($projectId);
        return $projectsAll;
    }

    /**
     * @param $projectId
     * @return bool
     */
    public function checkProjectHasTemplateParentRecursive($projectId){
        $projectId = !$projectId ? 0 : $projectId;
        $projectType = InstanceProvider::getRepository()->getProject()->checkProjectHasTemplateParentRecursive($projectId);
        return empty($projectType) ? false : true;
    }

    public function productProjectHasAllowReplicationLts($projectId) {
        /** @var Entities\Project $project */
        if($projectId) {
            $project = InstanceProvider::getRepository()->getProject()->find($projectId);
            $productProject = $this->getProductProject($project);
            if($productProject) {
                return $productProject->getAllowReplicationLts();
            }
            return null;
        }
        return null;
    }

    public function projectFinishPlanning($projectId)
    {
        /** @var Entities\Project $project */
        $project = InstanceProvider::getRepository()->getProject()->find($projectId);
        $this->finishPlanning($project);
    }

    public function finishPlanning(Entities\Project $project, &$projectNamesInCalc = array())
    {
        /** @var \Zeedhi\DevOps\Model\Projectmanagement\Repositories\ProjectRelationship $projectRelationshipRepository */
        $projectRelationshipRepository = InstanceProvider::getRepository()->getProjectRelationship();

        $this->environment->setDOCurrentUserIdByCurrentPMUser();

        if (!Utilities::getProjectTypeTimeLine()) {
            throw new\Exception('Não existe tipo de projeto com o controle "Baseline"!');
        }

        if ($project->getInCalc() == 'Y' && $project->getProjectType() && $project->getProjectType()->getUsetimeplanning() == 'Y') {
            $projectNamesInCalc[] = $project->getName();
        }
        if ($project->getParent()) {
            if ($project->getProjectType() && $project->getProjectType()->getExternaltype() != Entities\ProjectType::TYPE_PROJECT_CONTROL_PRODUCT) {
                $this->finishPlanning($project->getParent(), $projectNamesInCalc);
            }
        }

        $this->validateProjectCalc($projectNamesInCalc);

        $hasBaselineProject = $projectRelationshipRepository->findOneBy(array('project' => $project->getId(), 'type' => Entities\ProjectRelationship::TYPE_TIME_LINE));

        $connection = $this->entityManager->getConnection();
        try {
            $connection->beginTransaction();
            if ($project->getProjectType() &&
                $project->getProjectType()->getExternaltype() != Entities\ProjectType::TYPE_PROJECT_CONTROL_PRODUCT &&
                empty($hasBaselineProject)
            ) {
                $cloneProject = $this->saveProjectTimeLine($project);
                $allTasks = $this->cloneTasks($project, $cloneProject);
                $taskRelationshipFactory = InstanceProvider::getFactory()->getTaskRelationship();
                foreach ($allTasks as $task) {
                    /** @var Entities\Task $originTask */
                    $originTask = $task['origin'];
                    /** @var Entities\Task $copyTask */
                    $copyTask = $task['copy'];
                    $relationship = $taskRelationshipFactory->createTaskRelationship($originTask, $copyTask,
                        Entities\TaskRelationship::BASELINE_TYPE);
                    $this->entityManager->persist($relationship);
                }
                InstanceProvider::getService()->getProjectRelationship()->createProjectRelationship($project, $cloneProject,
                    Entities\ProjectRelationship::TYPE_TIME_LINE);
                $cloneProject->setTimePlanned($project->getTimePlanned());
                $this->entityManager->persist($cloneProject);
            }

            $project->setProjectStatus(Utilities::getProjectStatusProgress());
            $this->entityManager->persist($project);
            $this->entityManager->flush();
            $this->setsDevOps($project);
            $connection->commit();
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }

    }

    public function setParentInCloneTask($allTask,Entities\Task $targetTask,Entities\Task $cloneTask){
        if ($targetTask->getParent()) {
            foreach ($allTask as $task) {
                /** @var Entities\Task $originTask */
                $originTask = $task['origin'];
                /** @var Entities\Task $copyTask */
                $copyTask = $task['copy'];
                if($targetTask->getParent() && $targetTask->getParent()->getId() == $originTask->getId()){
                    $cloneTask->setParent($copyTask);
                    $this->entityManager->persist($cloneTask);
                    break;
                }
            }
        }
    }

    public function cloneTasks($project,$targetProject){
        $allTask  = array();
        $tasks = InstanceProvider::getRepository()->getTask()->findBy(array('project' => $project,'status' => Utilities::ACTIVE,'isIssue' => 'N'));
        $taskService = InstanceProvider::getService()->getTask();
        /** @var Entities\Task $task */
        foreach ($tasks as $task) {
            $copiedTask = $taskService->createTask($targetProject, $task->getLatency(),
                $task->getName(), $task->getEstimatedBeginDate(),
                $task->getEstimatedEndDate(), $task->getStatus(), $task->getDescription(),
                $task->getPercentCompleted(), $task->getDuration(), null,
                $this->getStatusStandard(), $task->getDateExceptionType(), $task->getDateException(), $task->getIsParent(), $task->getUrlToTest(),
                $task->getClassification());
            $copiedTask->setResponse($task->getResponse());
            $copiedTask->setTimePlanned($task->getTimePlanned());
            $copiedTask->setProgress($task->getProgress());
            //$this->copyTaskAttachments($task, $copiedTask, '');
            $allocationService = InstanceProvider::getService()->getAllocation();
            /** @var Entities\Allocation[] $allocations */
            $allocations = $allocationService->repositoryAllocation->findBy(array('task' => $task));
            foreach ($allocations as $allocation) {
                $allocationService->createAllocation($copiedTask, $allocation->getResource(),null,$copiedTask->getTaskStatus(),true);
            }

            $allTask[] = array('copy'=>$copiedTask,'origin' =>$task );
            $this->entityManager->flush();
        }

        foreach ($allTask as $task) {
            /** @var Entities\Task $originTask */
            $originTask = $task['origin'];

            /** @var Entities\Task $copyTask */
            $copyTask = $task['copy'];
            $this->setParentInCloneTask($allTask, $originTask,$copyTask);
        }
        return $allTask;
    }

    public function getAppropriationByProject($projectId)
    {
        $projectsAll = InstanceProvider::getRepository()->getProject()->getAppropriationByProject($projectId);
        return $projectsAll;
    }



    public function createProject($name, $parent, $status, $duration, \DateTime $endDate, \DateTime $beginDate, $manager,
                                  $sponsor, \Zeedhi\DevOps\Model\Projectmanagement\Entities\Structure $group, $dateExceptionType,
                                  $dateException = null, $projectType, $category, $description,
                                   \Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskStatus $projectStatus = null, $productVersion = null, $isLtsReplication = 'Y')
    {

        $project = InstanceProvider::getFactory()->getProject()->createProject($name, $parent, $status, $duration, $endDate, $beginDate, $group,
                                                                                $dateExceptionType, $dateException, $description, null,
                                                                                $projectType, $category, $projectStatus, $productVersion, $isLtsReplication);
        $this->entityManager->persist($project);
        $this->entityManager->flush();

        return $project;
    }

    public function changeMnemonicName($id, $model, $oldName, $newName)
    {

        $projectRepository = InstanceProvider::getRepository()->getProject();
        $projectArray = $projectRepository->getRelatedProjectFromModel($id, $model);
        $connection = $this->entityManager->getConnection();
        try {
            $oldName = '['. $oldName . ']';
            $connection->beginTransaction();
            foreach($projectArray as $project){
                $projectName = $projectRepository->getModelProjectName($project['ID']);

                $name = str_replace($oldName, $newName , $projectName[0]);
                $entity = InstanceProvider::getRepository()->getProject()->find($project['ID']);
                $project = InstanceProvider::getFactory()->getProject()->UpdateName($entity, $name['NAME']);
                $this->entityManager->persist($project);
                $this->setsDevOps($project);
                $this->entityManager->flush();
            }
            $connection->commit();
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
    }

    public function deleteProject($projectId)
    {
        /** @var Entities\Project $project */
        $connection = $this->entityManager->getConnection();
        try {
            $connection->beginTransaction();
            /** @var Entities\Project $project */
            $project = InstanceProvider::getRepository()->getProject()->find($projectId);
            $this->deleteProjectModel($project);
            $this->deleteAllSubProject($project);
            $this->deleteDevOpsAssociation($project);
            $project->setStatus(LogicalDeleted::STATUS_LOGICAL_DELETED);
            $this->entityManager->persist($project);
            $this->entityManager->flush();

            $connection->commit();
        }  catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
    }

    private function deleteAllSubProject(Entities\Project $project)
    {
        /** @var Entities\Project[] $subProjects */
        $subProjects = InstanceProvider::getRepository()->getProject()->findBy(array("parent" => $project));
        /** @var Entities\Task[] $tasks */
        $tasks = InstanceProvider::getRepository()->getTask()->findBy(array("project" => $project));
        foreach ($tasks as $task) {
            $appropriations = $this->getAppropriationByProject($project->getId());
            if(empty($appropriations)) {
                $task->setStatus('D');
                $this->entityManager->persist($task);
            }else{
                throw new \Exception(Messages::ERR_NOT_SUCCESSFUL_PROJECT_DELETED);
            }
        }
        foreach ($subProjects as $subProject) {
            $subProject->setStatus('D');
            $this->deleteDevOpsAssociation($subProject);
            $this->entityManager->persist($subProject);
            $this->deleteAllSubProject($subProject);
        }
    }

    public function deleteDevOpsAssociation(Entities\Project $project)
    {
        $entityToExclude = InstanceProvider::getRepository()->getProduct()->findOneBy(array('project' => $project));
        $typeEntityToExclude = self::TYPE_PRODUCT;
        if (empty($entityToExclude)) {
            $entityToExclude = InstanceProvider::getRepository()->getRelease()->findOneBy(array('project' => $project));
            $typeEntityToExclude = self::TYPE_RELEASE;
            if (empty($entityToExclude)) {
                $entityToExclude = InstanceProvider::getRepository()->getSubrelease()->findBy(array('project' => $project));
                $typeEntityToExclude = self::TYPE_SUBRELEASE;
            }
        }

        if (!empty($entityToExclude)) {
            $data = array();

            if(is_array($entityToExclude)){
                foreach($entityToExclude as $entity){
                    $auxArray = array(
                        "requestType" => "Row",
                        "row" => array(
                            "ID" => $entity->getId()
                        )
                    );
                    array_push($data, $auxArray);
                }
            } else {
                $data = array(
                    "requestType" => "Row",
                    "row" => array(
                        "ID" => $entityToExclude->getId()
                    )
                );
            }
            $token = $this->getTokenInDevOps();
            if ($typeEntityToExclude == self::TYPE_PRODUCT) {
                $this->deleteProductInDevOps($token, $data);
            } else if ($typeEntityToExclude == self::TYPE_RELEASE) {
                $this->deleteReleaseInDevOps($token, $data);
            } else if ($typeEntityToExclude == self::TYPE_SUBRELEASE) {
                if(is_array($data)){
                    foreach($data as $subRelease){
                        $this->deleteSubreleaseInDevOps($token, $subRelease);
                    }
                } else{
                    $this->deleteSubreleaseInDevOps($token, $data);
                }
            }
        }
    }

    public function getTokenInDevOps() {
        if($this->environment->getTokenDevOps()) {
            $token = $this->environment->getTokenDevOps();
        } else {
            $param = array(
                'CLIENT_ID' => $this->environment->CLIENT_ID,
                'CLIENT_SECRET' => $this->environment->PRIVATE_KEY_DEVOPS
            );
            $token = $this->environment->callWebService(null, '/do_requestToken', $param, 'token',
                $this->environment->PRIVATE_KEY_DEVOPS, $this->environment->DEVOPS_URL);
            $token = $token[0];
            $this->environment->setTokenDevOps($token);
        }

        return $token;
    }

    public function deleteProductInDevOps($token, $row) {
        $this->environment->callWebService($token, '/product/remove', $row, 'product',
            $this->environment->PRIVATE_KEY_DEVOPS, $this->environment->DEVOPS_URL);
    }

    public function deleteReleaseInDevOps($token, $row) {
        $this->environment->callWebService($token, '/release_query/remove', $row, 'release',
            $this->environment->PRIVATE_KEY_DEVOPS, $this->environment->DEVOPS_URL);
    }

    public function deleteSubreleaseInDevOps($token, $row) {
        $this->environment->callWebService($token, '/subrelease_query/remove', $row, 'subrelease',
            $this->environment->PRIVATE_KEY_DEVOPS, $this->environment->DEVOPS_URL);
    }

    public function saveProductInDevOps($token, $row) {
        $this->environment->callWebService($token, '/product/save', $row, 'product',
            $this->environment->PRIVATE_KEY_DEVOPS, $this->environment->DEVOPS_URL);
    }

    public function saveReleaseInDevOps($token, $row) {
        $this->environment->callWebService($token, '/saveRelease/save', $row, 'release',
            $this->environment->PRIVATE_KEY_DEVOPS, $this->environment->DEVOPS_URL);
    }

    public function removePathReleaseInDevOps($token, $filter) {
        $this->environment->callWebService($token, '/removePathRelease', $filter, 'release',
            $this->environment->PRIVATE_KEY_DEVOPS, $this->environment->DEVOPS_URL);
    }
    public function removePathSubReleaseInDevOps($token, $filter) {
        $this->environment->callWebService($token, '/removePathSubRelease', $filter, 'subRelease',
            $this->environment->PRIVATE_KEY_DEVOPS, $this->environment->DEVOPS_URL);
    }

    public function saveSubreleaseInDevOps($token, $row) {
        $this->environment->callWebService($token, '/subrelease/save', $row, 'subrelease',
            $this->environment->PRIVATE_KEY_DEVOPS, $this->environment->DEVOPS_URL);
    }

    private function activeProductInDevOps(Product $product = null,$isActive) {
        if($product){
            $product->setIsActive($isActive);
            $this->entityManager->persist($product);
            $project = $product->getProject();
            if($project){
                /** @var Release $release */
                $release = InstanceProvider::getRepository()->getRelease()->findOneBy(array('project' => $project));
                $release->setIsActive($isActive);
                $this->entityManager->persist($release);
            }
            $this->entityManager->flush();
        }
    }

    private function activeReleaseInDevOps(Release $release = null,$isActive) {
        if($release){
            $release->setIsActive($isActive);
            $this->entityManager->persist($release);
            $this->entityManager->flush();
        }
    }

    private function activeSubreleaseInDevOps(Subrelease $subrelease = null,$isActive) {
        if($subrelease){
            $subrelease->setIsActive($isActive);
            $this->entityManager->persist($subrelease);
            $this->entityManager->flush();
        }
    }


    protected function getUrlsProject($projectId)
    {
        $projectRepositories = array();
        /** @var Entities\Repository[] $targetProjectRepositories */
        $targetProjectRepositories = InstanceProvider::getRepository()->getRepository()->findBy(array('project' => $projectId));
        foreach ($targetProjectRepositories as $projectRepository) {
            if (!in_array($projectRepository->getUrl(), $projectRepositories)) {
                array_push($projectRepositories, $projectRepository->getUrl());
            }
        }
        return $projectRepositories;
    }

    /**
     * @param $projectModelFirst
     * @param $project
     * @return array
     */
    protected function getTagNameModel(Entities\Project $projectModelFirst,Entities\Project $project)
    {
        $tagBegin = '';
        $tagEnd = '';
        $projectModelName = $projectModelFirst->getName();
        $projectName = $project->getName();
        $tag = explode($projectModelName,$projectName);
        if($projectModelName == $projectName){
            $productProject = $this->getProductProject($project);
            if($productProject){
                $tagBegin = $productProject->getName().' ';
                $project->setName($tagBegin.$projectName);
            }
        }else if(count($tag) == 2){
            $tagBegin = $tag[0];
            $tagEnd = $tag[1];
        }
        return array($tagBegin, $tagEnd);
    }

    /**
     * @param Entities\Project $project
     * @return Entities\Project
     */
    public function saveProjectTimeLine(Entities\Project $project)
    {
        $projectType = Utilities::getProjectTypeTimeLine();
        $name = $project->getName();
        $parent = $project->getParent();
        $status = $project->getStatus();
        $duration = $project->getDuration();
        $endDate = $project->getEndDate();
        $beginDate = $project->getBeginDate();
        $group = $project->getStructure();
        $dateExceptionType = $project->getDateExceptionType();
        $dateException = $project->getDateException();
        $description = $project->getDescription();
        $isBoard = $project->getIsIssueProject();
        $category = $project->getClassification();
        $projectStatus = $project->getProjectStatus();
        $isLtsReplication = $project->getIsLtsReplication();
        $cloneProject = InstanceProvider::getFactory()->getProject()->createProject($name, $parent, $status, $duration, $endDate, $beginDate,
            $group, $dateExceptionType, $dateException, $description, $isBoard, $projectType, $category, $projectStatus, null, $isLtsReplication);
        $cloneProject->setTimePlanned($project->getTimePlanned());
        $this->entityManager->persist($cloneProject);
        $this->entityManager->flush();


       // $parentRepositories = InstanceProvider::getRepository()->getRepository()->findBy(array('project' => $project));
        $parentResourceProject = InstanceProvider::getRepository()->getResourceProject()->findBy(array('project' => $project));

       // $this->cloneProjectRepositories($cloneProject, $parentRepositories, $parentResourceProject);
        $this->cloneResourceProject($cloneProject, $parentResourceProject,true);

       // $this->copyProjectAttachments($project,$cloneProject);

        return $cloneProject;
    }

    /**
     * @param $projectNamesInCalc
     * @throws \Exception
     */
    public function validateProjectCalc(&$projectNamesInCalc)
    {
        if (!empty($projectNamesInCalc)) {

            $message = 'Para finalizar o planejamento é necessário que o(s) projeto(s) abaixo tenha(m) as datas calculadas: <BR><BR>';
            foreach ($projectNamesInCalc as $projectNameInCalc) {
                $message .= '"' . $projectNameInCalc . '"<BR><BR>';
            }

            throw new \Exception($message);
        }
    }

    private function buildProductRow(Product $product, $isActive,$userIds)
    {
        $dataSet = array(
            "requestType" => "DataSet",
            "dataset" => array ( array(
                'ID' => $product->getId(),
                'NAME' => $product->getName(),
                'DESCRIPTION' => $product->getDescription(),
                'BASE_URL' => $product->getBaseUrl(),
                'IS_ACTIVE' => $isActive,
                'USER_ID' => $userIds,
                'VERSION' => $product->getVersion(),
                'REVISION' => $product->getRevision(),
                'BUILD' => $product->getBuild(),
                'PROJECT_ID' => $product->getProject()->getId(),
                'isMessage' => false,
                '__is_new' => false,
                'changed' => true
            ) )
        );

        return $dataSet;
    }

    private function buildReleaseRow(Release $release, $isActive, $userIds)
    {
        $dataSet = array(
            "requestType" => "DataSet",
            "dataset" => array ( array(
                'ID' => $release->getId(),
                'NAME' => $release->getName(),
                'PRODUCT_ID' => $release->getProduct() ? $release->getProduct()->getId() : 0,
                'USER_ID' => $userIds,
                'PRODUCT_SERVER_ID' => $release->getProductServer() ? $release->getProductServer()->getId() : 0,
                'UPDATE_VERSION' => $release->getUpdateVersion(),
                'IS_FIX_RELEASE' => $release->getIsFixRelease(),
                'STATUS' => $release->getStatus(),
                'VERSION' => $release->getVersion(),
                'REVISION' => $release->getRevision(),
                'BUILD' => $release->getBuild(),
                'DESCRIPTION' => $release->getDescription(),
                'IS_ACTIVE' => $isActive,
                'isMessage' => false,
                '__is_new' => false,
                'changed' => true
            ) )
        );

        return $dataSet;
    }

    private function buildSubreleaseRow(Subrelease $subrelease, $isActive, $userIds)
    {
        $dataSet = array(
            "requestType" => "DataSet",
            "dataset" => array ( array(
                'ID' => $subrelease->getId(),
                'RELEASE_ID' => $subrelease->getRelease()->getId(),
                'USER_ID' => $userIds,
                'NAME' => $subrelease->getName(),
                'DESCRIPTION' => $subrelease->getDescription(),
                'IS_ACTIVE' => $isActive,
                'STATUS' => $subrelease->getStatus(),
                'URL_TO_TEST' => $subrelease->getUrlToTest(),
                'BRANCH_TO_DEPLOY' => $subrelease->getBranchToDeploy(),
                'PRODUCT_SERVER_ID' => $subrelease->getProductServer() ? $subrelease->getProductServer()->getId() : 0,
                'PATH_TO_TEST' => $subrelease->getPathToTest(),
                'VERSION' => $subrelease->getVersion(),
                'REVISION' => $subrelease->getRevision(),
                'BUILD' => $subrelease->getBuild(),
                'PROJECT_ID' => $subrelease->getProject()->getId(),
                'isMessage' => false,
                '__is_new' => false,
                'changed' => true
            ) )
        );

        return $dataSet;
    }

    public function updateProject($entity, $name, $beginDate, $endDate, $status, $structureProject,
                                  $parentProject, $dateExceptionType = null, $dateException = null, $description = null,
                                  $projectType = null, $category = null, $projectStatus = null, $productVersionDevId = null, $isLtsReplication = 'Y'
    )
    {
        $duration = InstanceProvider::getService()->getCalendarTakCalendar()->setDuration($entity);
        $productVersion = null;
        if($productVersionDevId){
            $productVersion = InstanceProvider::getRepository()->getProductVersion()->find($productVersionDevId);
        }
        \Util\InstanceProvider::getLog()->setLog('updateProject setDuration');
        $entity = InstanceProvider::getFactory()->getProject()->updateProject($entity, $name, $beginDate, $endDate, $status, $structureProject,
            $parentProject, $duration, $dateExceptionType, $dateException, $description, $projectType, $category, $projectStatus,
            $productVersion, $isLtsReplication
        );

        \Util\InstanceProvider::getLog()->setLog('updateProject updateProject');
        $this->setStatusAllTasksProject($entity, $status);
        \Util\InstanceProvider::getLog()->setLog('updateProject setStatusAllTasksProject');
        $this->entityManager->persist($entity);
        \Util\InstanceProvider::getLog()->setLog('updateProject persist');
        $this->entityManager->flush();
        \Util\InstanceProvider::getLog()->setLog('updateProject flush');

        return $entity;
    }

    /**
     * @param Entities\Project $project
     * @return null|Entities\Project
     */

    public function getModelProjectParent(Entities\Project $project = null){
        return \Zeedhi\DevOps\Util\Utilities::getModelProjectParent($project);
    }

    public function getDataSetForEditProjectBasic($projectId)
    {
        \Util\InstanceProvider::getLog()->setLog('',true);
        $projectTypeIdsArray = explode(',', $this->projectTypeIds);
        $projectRepository = InstanceProvider::getRepository()->getProject();
        $projectArray = $projectRepository
            ->getDataSetForEditProjectBasic($projectId, $this->environment->getCurrentOrganizationId());
        \Util\InstanceProvider::getLog()->setLog('select');
        foreach ($projectArray as &$project) {
            if (!empty($project['DATE_EXCEPTION_ABBREVIATION'])) {
                $project['DATE_EXCEPTION_TYPE'] = $project['DATE_EXCEPTION_ABBREVIATION'];
                $project['DATE_EXCEPTION_NAME'] = self::$DATA_SET_EXCEPTION_LEAF[$project['DATE_EXCEPTION_ABBREVIATION']];
            }
            $project['IS_PROJECT_TYPE_ALLOWED'] = in_array($project['PROJECT_TYPE_ID'], $projectTypeIdsArray);
            $dateBeginMax = null;
            $dateEndMin = null;
            $project['OLD_END'] = $project['END'];
            $project['OLD_START'] = $project['START'];
            $project['dateBeginMax'] = null;
            $project['dateEndMin'] = null;
            $project['TIME_PLANNED'] = $project['TIME_PLANNED'] ? $project['TIME_PLANNED'] : null;
            \Util\InstanceProvider::getLog()->setLog('monte de coisas rapidas');
            /** @var Entities\Project $projectEntity */
            $projectEntity = $projectRepository->find($project['ID']);
            $projectRelationshipService = InstanceProvider::getService()->getProjectRelationship();
            $modelProject = $projectRelationshipService->getOneProjectRelationship($projectEntity,
                Entities\ProjectRelationship::TYPE_MODEL);

            $modelProjectPrime = null;
            if($modelProject) {
                $modelProjectPrime = $projectRelationshipService->getOneProjectRelationship($projectEntity,
                    Entities\ProjectRelationship::TYPE_MODEL_GENERATION);
            }
            if($modelProjectPrime){
                $project['MODELS'] = $modelProjectPrime->getId();
                $project['MODELS_NAME'] = $modelProjectPrime->getName();
                $project['MODELS_IS_RESTRICTED'] = $modelProjectPrime->getProjectType()->getExternaltype() == Entities\ProjectType::TYPE_PROJECT_CONTROL_MODEL_RESTRICTED;
                $project['MODELS_IS_FIRST'] = $modelProjectPrime->getId() == $modelProject->getParent()->getId();

            }else {
                $modelProjectPrime = $this->getModelProjectParent($projectEntity);
                $project['IS_MODEL'] = !!$modelProjectPrime;
                if ($modelProjectPrime
                    && $modelProjectPrime->getProjectType()
                    && $modelProjectPrime->getProjectType()->getExternaltype() == Entities\ProjectType::TYPE_PROJECT_CONTROL_MODEL_RESTRICTED
                ) {
                    $project['MODELS_HAS_RELATIONSHIP'] = $projectRelationshipService->hasProject($projectEntity, Entities\ProjectRelationship::TYPE_MODEL);
                }
            }
            \Util\InstanceProvider::getLog()->setLog('modelos');
            if($project['USETIMEPLANNING'] == 'Y'){
                $project['TOTAL_TIME'] = InstanceProvider::getService()->getCalendarTakCalendar()->getDurationInDays($projectEntity,
                    $projectEntity->getEstimatedBeginDate(), $projectEntity->getEstimatedEndDate());
            }else{
                $project['TOTAL_TIME'] = Utilities::secondsToTime($project['TOTAL_TIME']);
            }
            \Util\InstanceProvider::getLog()->setLog('duração segunto para dias');
        }
        return $projectArray;
    }

    public function getDatesMinAndMaxByProjectId($projectId)
    {
        /** @var Entities\Project $project */
        $project = InstanceProvider::getRepository()->getProject()->find($projectId);
        $dateBeginMax = null;
        $dateEndMin = null;
        if ($project && $project->getProjectType() && $project->getProjectType()->getUsetimeplanning() == 'Y') {
            list($dateBeginMax, $dateEndMin) = $this->getDatesRestrictionsMinAndMax($projectId);
        }
        $return = array();
        $return['dateBeginMax'] = $dateBeginMax ? date_format($dateBeginMax, 'd/m/Y') : $dateBeginMax;
        $return['dateEndMin'] = $dateEndMin ? date_format($dateEndMin, 'd/m/Y') : $dateEndMin;
        return array($return);

    }
    public function getDataSetForEditProjectException($project)
    {
        return $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\Commitment')
            ->getDataSetForEditProjectException($project, $this->environment->getCurrentOrganizationId());
    }

    public function getDataSetForEditProjectCalendar($project)
    {
        return InstanceProvider::getRepository()->getProject()
            ->getDataSetForEditProjectCalendar($project, $this->environment->getCurrentOrganizationId());
    }

    public function getDataSetForEditProjectAdvanced($project)
    {
        return $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\Repository')->getProjectRepositories($project);
    }

    public function findModelProject($projectParentId,$filter = array())
    {
        $organizationId = $this->environment->getCurrentOrganizationId();
        $userId = $this->environment->getCurrentUserId();
        $filterString = Utilities::createQueryFiltered($filter);
        $hasProduct = false;
        $projectRepository = InstanceProvider::getRepository()->getProject();
        if($projectParentId){
            /** @var Entities\Project $projectParent */
            $projectParent = $projectRepository->findOneBy(array('id' => $projectParentId));
            $hasProduct =  !!$this->getProductOrModelProject($projectParent);
        }
        $projects = $projectRepository->getModelProject($organizationId,$userId,$filterString);
        $profiles = $this->environment->getCurrentUserProfiles();
        while(!$profiles){
            $profiles = InstanceProvider::getService()->getResource()->getProfileByCurrentUser();
        }

        $result = array();
        foreach ($projects as $project) {
            $projectParentIds = DevOpsUtilities::getAllProjectChildrenIdInArray($project['ID']);
            /** @var Entities\Project[] $projectParents */
            $projectParents = $projectRepository->findBy(array('id' => $projectParentIds ));
            $validate = true;
            $hasProductParents = false;
            $hasReleaseOrSubRelease = false;
            $projectTypeDevOps = self::arrayProjectTypeDevOps();
            $projectTypeProductAndModel = array(
                Entities\ProjectType::TYPE_PROJECT_CONTROL_PRODUCT,
                Entities\ProjectType::TYPE_PROJECT_CONTROL_MODEL_FREE,
                Entities\ProjectType::TYPE_PROJECT_CONTROL_MODEL_RESTRICTED,
            );
            foreach ($projectParents as $projectParent) {
                $projectType = $projectParent->getProjectType();
                if(!$hasProduct){
                    if(in_array($projectType->getExternaltype(),$projectTypeDevOps)){
                        $hasReleaseOrSubRelease = true;
                    }

                    if(in_array($projectType->getExternaltype(),$projectTypeProductAndModel)){
                        $hasProductParents = true;
                    }
                }
            }

            if($hasReleaseOrSubRelease && !$hasProductParents){
                $validate = false;
            }

            if($validate){
                $result[] = $project;
            }
        }
        return $result;
    }

    public function findModelProjectForFilter()
    {
        $organizationId = $this->environment->getCurrentOrganizationId();
        $userId = $this->environment->getCurrentUserId();
        $projects = InstanceProvider::getRepository()->getProject()->getModelProject($organizationId,$userId);
        return $projects;
    }

    public function getBasicModelProject($projectModelId)
    {
        $projectRepository = InstanceProvider::getRepository()->getProject();
        $projectModel = $projectRepository->findOneBy(array('parent' => $projectModelId));
        if(empty($projectModel)){
            throw new \Exception(Messages::ERR_NOT_MODEL_CHILDREN,1);
        }
        $projectArray = $projectRepository->getDataSetForEditProjectBasic($projectModel->getId(),
            $this->environment->getCurrentOrganizationId());
        foreach ($projectArray as &$project) {
            if (!empty($project['DATE_EXCEPTION_ABBREVIATION'])) {
                $project['DATE_EXCEPTION_TYPE'] = $project['DATE_EXCEPTION_ABBREVIATION'];
                $project['DATE_EXCEPTION_NAME'] = self::$DATA_SET_EXCEPTION_LEAF[$project['DATE_EXCEPTION_ABBREVIATION']];
            }
            $dateBeginMax = null;
            $dateEndMin = null;
            $project['OLD_END'] = $project['END'];
            $project['OLD_START'] = $project['START'];
            $project['dateBeginMax'] = null;
            $project['dateEndMin'] = null;
            /** @var Entities\Project $projectEntity */
            $projectEntity = $projectRepository->find($project['ID']);
            $project['TOTAL_TIME'] = InstanceProvider::getService()->getCalendarTakCalendar()->getDurationInDays($projectEntity,
                $projectEntity->getEstimatedBeginDate(), $projectEntity->getEstimatedEndDate());
            $project['ID'] = null;
            $project['CREATED_BY'] = null;
            $project['CREATED_AT'] = null;
            $project['UPPER_PROJECT'] = null;
            $project['UPPER_PROJECT_NAME'] = null;
            if ($project['PROJECT_STATUS_CONTROL'] == "D") {
                $projectStatusPlanning = Utilities::getProjectStatusPlanning();
                $project['PROJECT_STATUS_CONTROL'] = $projectStatusPlanning->getControl();
                $project['PROJECT_STATUS'] = $projectStatusPlanning->getLabel();
                $project['PROJECT_STATUS_ID'] = $projectStatusPlanning->getId();

            }
        }
        return $projectArray;
    }


 public function getHeadSprintProject($projectModelId)
    {
        $projectRepository = InstanceProvider::getRepository()->getProject();

        $projectArray = $projectRepository->getHeadSprintProject($projectModelId);
        if(empty($projectArray)){
            throw new \Exception(Messages::ERR_NOT_SPRINT_PROJECT,1);
        }

        return $projectArray;
    }



    public function getTailSprintProject($projectModelId)
    {
        $projectRepository = InstanceProvider::getRepository()->getProject();

        $projectArray = $projectRepository->getTailSprintProject($projectModelId);


        return $projectArray;
    }

    public function validateRelationDevOps(Entities\Project $project)
    {
        $validate = false;
        if(InstanceProvider::getRepository()->getProject()->validateRelationDevOps($project->getId())){
            $validate = true;
        }
        return $validate;
    }

    public function validateProjectStatus(Entities\Project $project)
    {

        $validate = false;
        if ($project->getProjectStatus() && $project->getProjectStatus()->getControl() != self::PROJECT_NOT_INITIALIZED_STATUS) {
            $validate = true;
        } else if ($this->validateRelationDevOps($project)) {
            $validate = true;
        }
        return $validate;
    }

    public function removePathRelease(Release $release = null)
    {
        if($release
            && $release->getProject()
            && $release->getProject()->getProjectStatus()
            && $release->getProject()->getProjectStatus()->getControl() == Entities\TaskStatus::TASK_STATUS_CONTROL_CANCELED){
            $filter = $this->buildRemovePathReleaseFilter($release);
            $this->removePathReleaseInDevOps($this->getTokenInDevOps(), $filter);
        }
    }

    public function removePathSubRelease(SubRelease $subRelease = null)
    {
        if($subRelease
            && $subRelease->getProject()
            && $subRelease->getProject()->getProjectStatus()
            && $subRelease->getProject()->getProjectStatus()->getControl() == Entities\TaskStatus::TASK_STATUS_CONTROL_CANCELED){
            $filter = $this->buildRemovePathSubReleaseFilter($subRelease);
            $this->removePathSubReleaseInDevOps($this->getTokenInDevOps(), $filter);
        }
    }

    public function buildRemovePathReleaseFilter(Release $release)
    {
        $dataSet = array(
            "requestType" => "FilterData",
            "filter" => array(array(
                'value' => $release->getId(),
                'operator' => '=',
                'name' => 'RELEASE_ID'
            ))
        );

        return $dataSet;
    }

    public function buildRemovePathSubReleaseFilter(SubRelease $subRelease)
    {
        $dataSet = array(
            "requestType" => "FilterData",
            "filter" => array(array(
                'value' => $subRelease->getId(),
                'operator' => '=',
                'name' => 'SUBRELEASE_ID'
            ))
        );

        return $dataSet;
    }

    public function setsDevOps(Entities\Project $project)
    {
        try {
            $this->environment->setDOCurrentUserIdByCurrentPMUser();
            $externalType = $project->getProjectType()->getExternaltype();
            $managers = InstanceProvider::getService()->getResourceProject()->getManagersByProject($project);
            $DOManagers = Utilities::getZhUserByResource($managers);
            $userIds = Utilities::getArrayByArrayObjectWithGet($DOManagers,array('getId'));
            if ($externalType && $this->validateProjectStatus($project)) {
                if ($externalType == self::PROJECT_TYPE_PRODUCT) {
                    $product = $this->setDevOpsWhenProjectTypeProduct($project, $DOManagers);
                    $this->activeProductInDevOps($product,$project->getStatusForDevOps());
                } else if ($externalType == self::PROJECT_TYPE_RELEASE) {
                    $release = $this->setDevOpsWhenProjectTypeRelease($project, $DOManagers);
                    $this->activeReleaseInDevOps($release, $project->getStatusForDevOps());
                    $this->removePathRelease($release);
                } else if ($externalType == self::PROJECT_TYPE_SUB_RELEASE) {
                    $subrelease = $this->setDevOpsWhenProjectTypeSubRelease($project, $DOManagers);
                    $this->activeSubreleaseInDevOps($subrelease, $project->getStatusForDevOps());
                    $this->removePathSubRelease($subrelease);

                    /** @var Release $release */
                    $release = InstanceProvider::getRepository()->getRelease()->findOneBy(array('project' => $project));
                    if(!empty($release)) {
                        $this->activeReleaseInDevOps($release, $project->getStatusForDevOps());
                        $this->removePathRelease($release);
                    }
                } else if ($externalType == self::PROJECT_TYPE_ERROR) {
                    if($project->getProductVersion()){
                        $productVersionRepository = InstanceProvider::getRepository()->getProductVersion();
                        $productRepository = InstanceProvider::getRepository()->getProduct();

                        $organizationId = $this->environment->getCurrentOrganizationId();
                        $projectProduct = $this->getProductProject($project);
                        $productId = $productRepository->findOneBy(array('project' => $projectProduct->getId()));
                        $version = $project->getProductVersion()->getVersion();
                        $revision = $project->getProductVersion()->getRevision();
                        $productVersionId = $project->getProductVersion()->getId();

                        $this->setDevOpsErrorTypeProjectSetSubRelease($project, $DOManagers, $userIds, $project->getProductVersion(), ' - [' . $version . '.' . $revision . ']');
                        if($project->getProductVersion()->getIsLts() == DevopsEntities\ProductVersion::IS_LTS_YES && $project->getIsLtsReplication() === 'Y') {
                            $nextVersionsLts = $productVersionRepository->findNextVersionsLts($productId->getId(), $organizationId, $version, $revision, $productVersionId);
                            foreach ($nextVersionsLts as $version) {
                                $productVersion = $productVersionRepository->find($version['ID']);
                                $this->setDevOpsErrorTypeProjectSetSubRelease($project, $DOManagers, $userIds, $productVersion, ' - [' . $version['VERSION'] . ']');
                            }
                            $this->setDevOpsErrorTypeProjectSetSubRelease($project, $DOManagers, $userIds, null, ' - [STABLE]');
                        }
                    } else {
                        $this->setDevOpsErrorTypeProjectDefault($project, $DOManagers, $userIds);
                    }
                }
            }
        } catch (\Exception $e) {
            if($e->getCode() != 2){
                throw $e;
            }
        }
    }

    public function setDevOpsErrorTypeProjectDefault($project, $DOManagers, $userIds)
    {
        $subrelease = $this->setDevOpsWhenProjectTypeError($project, $DOManagers);
        $this->activeSubreleaseInDevOps($subrelease, $project->getStatusForDevOps(), $userIds);
        $this->removePathSubRelease($subrelease);
    }

    public function setDevOpsErrorTypeProjectSetSubRelease($project, $DOManagers, $userIds, $productVersion = null, $nameConcat = '')
    {
        $subrelease = $this->setDevOpsWhenProjectTypeError($project, $DOManagers, $productVersion, $nameConcat);
        $this->activeSubreleaseInDevOps($subrelease, $project->getStatusForDevOps(), $userIds);
        $this->removePathSubRelease($subrelease);
    }

    /**
     * @param Product $product
     * @param string $userType
     * @return null|\Zeedhi\DevOps\Model\Servermanager\Entities\ProductServer
     */
    public function getProductServerForRelease(Product $product, $userType = 'C')
    {
        $productServer = null;
        $productsServer = InstanceProvider::getRepository()->getProductServer()->findBy(array('usetype' => $userType, 'product' => $product, 'isActive' => Utilities::IS_ACTIVE_YES ));
        if (count($productsServer) == 1) {
            $productServer = $productsServer[0];
        }
        return $productServer;
    }

    /**
     * @param Entities\Project|null $project
     * @return Entities\Project|null
     */
    public static function getProductProject(Entities\Project $project = null)
    {
        $productProject = null;
        if ($project) {
            if ($project->getProjectType()->getExternaltype() == self::PROJECT_TYPE_PRODUCT) {
                $productProject = $project;
            } else {
                $productProject = self::getProductProject($project->getParent());
            }
        }
        return $productProject;
    }

    public static function getProductOrModelProject(Entities\Project $project = null)
    {
        $projectTypes = array(
            self::PROJECT_TYPE_PRODUCT,
            self::PROJECT_TYPE_MODEL_FREE,
            self::PROJECT_TYPE_MODEL_RESTRICTED,
        );
        $productProject = null;
        if ($project) {
            if (in_array($project->getProjectType()->getExternaltype(), $projectTypes)) {
                $productProject = $project;
            } else {
                $productProject = self::getProductOrModelProject($project->getParent());
            }
        }
        return $productProject;
    }

    public static function getReleaseProject(Entities\Project $project = null)
    {
        $productProject = null;
        if ($project) {
            $externalType = $project->getProjectType()->getExternaltype();
            if ($externalType != self::PROJECT_TYPE_PRODUCT) {
                if ($externalType == self::PROJECT_TYPE_RELEASE) {
                    $productProject = $project;
                } else {
                    $productProject = self::getReleaseProject($project->getParent());
                }
            }
        }
        return $productProject;
    }

    public function changeProjectType(Entities\Project $project,$projectType)
    {
        $changeProjectType = true;
        if($project->getProjectType() && $project->getProjectType()->getId() == $projectType){
            $changeProjectType = false;
            $product = InstanceProvider::getRepository()->getProduct()->findOneBy(array('project' => $project));
            $release = InstanceProvider::getRepository()->getRelease()->findOneBy(array('project' => $project));
            $subRelease = InstanceProvider::getRepository()->getSubrelease()->findOneBy(array('project' => $project));
            if($product || $release || $subRelease){
                $changeProjectType = true;
            }
        }
        return $changeProjectType;
    }

    public function changeProjectStatus(Entities\Project $project,$projectStatusNewId)
    {
        $changeProjectStatus = false;
        $projectStatus = $project->getProjectStatus();
        if($projectStatusNewId
            && $projectStatus
            && $projectStatus->getId() != $projectStatusNewId
            && $projectStatus->getControl() == self::PROJECT_NOT_INITIALIZED_STATUS ){
            $changeProjectStatus = true;
        }
        return $changeProjectStatus;
    }

    public function validateSetDevOps(Entities\Project $project,$projectType,$projectStatus)
    {
        $validate = false;
        if ($this->changeProjectType($project, $projectType) || $this->changeProjectStatus($project, $projectStatus)) {
            $validate = true;
        }
        return $validate;
    }


    /**
     * @param $projectStatus
     * @param $project
     * @return bool
     */
    private function validatePermissionChangeProjectStatus(Entities\TaskStatus $projectStatus = null, Entities\Project $project = null)
    {
        $validate = true;
        if ($project && $project->getProjectStatus() && $project->getProjectStatus()->getControl() == self::PROJECT_NOT_INITIALIZED_STATUS) {
            $profileIds = $this->environment->getCurrentUserProfiles();
            /** @var Entities\ProjectApproval[] $profilesApproval */
            $profilesApproval = InstanceProvider::getRepository()->getProjectApproval()->findBy(array('projectType' => $project->getProjectType()));
            $validate = false;
            if (!empty($profilesApproval)) {
                foreach ($profilesApproval as $profileApproval) {
                    if (in_array($profileApproval->getProfile(), $profileIds)) {
                        $validate = true;
                    }
                }
            } else {
                $validate = true;
            }
        }
        if ($project
            && $project->getProjectStatus()
            && in_array($project->getProjectStatus()->getControl(), self::arrayStatusNotModify())
        ) {
            $validate = false;
        }
        return $validate;
    }

    /**
     * @param $projectStatus
     * @param $project
     * @return bool
     */
    private function validateSetDevOpsParent(Entities\TaskStatus $projectStatus = null, Entities\Project $project = null)
    {
        return $projectStatus
        && $projectStatus->getControl() == self::PROJECT_IN_PROGRESS_STATUS
        && $project
        && $project->getProjectType()
        && $project->getProjectType()->getExternaltype() != self::PROJECT_TYPE_PRODUCT
        && $project->getProjectType()->getExternaltype() != self::PROJECT_TYPE_MODEL_RESTRICTED
        && $project->getProjectType()->getExternaltype() != self::PROJECT_TYPE_MODEL_FREE
        && $project->getProjectType()->getUsetimeplanning() != Entities\ProjectType::IS_USE_TIME_PLANNING
        && $project->getParent()
        && $this->validatePermissionChangeProjectStatus($projectStatus,$project->getParent());
    }

    public function setParentStatusProject(Entities\Project $project )
    {
        $projectStatus = $project->getProjectStatus();
        $parentProject = $project->getParent();
        if ($this->validateSetDevOpsParent($projectStatus, $project)) {
            $parentProject->setProjectStatus($projectStatus);
            $this->entityManager->persist($parentProject);
            $this->setParentStatusProject($parentProject);
            $this->setsDevOps($parentProject);
        }
    }

    public function saveProjectConfig($projectId,$notifyIssueOpening,$productType, $allowReplicationLts){
        /** @var Entities\Project $project */
        $project = InstanceProvider::getRepository()->getProject()->find($projectId);
        $project->setNotifyIssueOpening($notifyIssueOpening);
        $project->setProductType($productType);
        $project->setAllowReplicationLts($allowReplicationLts);
        $this->entityManager->persist($project);
        $this->entityManager->flush();
    }

    public function editProject($projectId, $name, $managers, $sponsor, $status, $group, $parent, $beginDate, $endDate,
                                $dateExceptionType = 'MB', $dateException = null, $mergeNewManager = array(), $mergeSponsor = true, $description = null,
                                $projectType = 1, $category, $projectStatus = null, $mergeOldManager = array(), $recursiveSponsor = false, $productVersionDevId = null,
                                $isLtsReplication = 'Y')
    {

        \Util\InstanceProvider::getLog()->setLog('',true);

        $connection = $this->entityManager->getConnection();
        try {
            $connection->beginTransaction();
            $projectRepository = InstanceProvider::getRepository()->getProject();
            /** @var Entities\Project $project */
            $project = $projectRepository->find($projectId);

            $setsDevOps = $this->validateSetDevOps($project,$projectType,$projectStatus);
            \Util\InstanceProvider::getLog()->setLog('setsDevOps');
            $changeProjectStatus = $this->changeProjectStatus($project,$projectStatus);
            \Util\InstanceProvider::getLog()->setLog('changeProjectStatus');
            if ($projectType) {
                $projectTypeRepository = InstanceProvider::getRepository()->getProjectType();
                $projectType = $projectTypeRepository->find($projectType);
                if ($projectType->getExternalType() === self::PROJECT_TYPE_ISSUE) {
                    $issueTypes = $projectTypeRepository
                        ->findBy(array('externaltype' => self::PROJECT_TYPE_ISSUE,
                            'organization' => $this->environment->getCurrentOrganizationId()));
                    foreach ($issueTypes as $issueType) {
                        $projectIssue = $projectRepository->findOneBy(array('projectType' => $issueType->getId(),
                            'parent' => $project->getParent()));
                        if (!empty($projectIssue) && $projectIssue->getId() !== $project->getId()) {
                            throw new \Exception('Já existe um projeto de Issue nesse produto');
                        }
                    }
                }
            }

            \Util\InstanceProvider::getLog()->setLog('verifica issue');

            $parentProject = null;
            $structureProject = null;
            if (!empty($parent)) {
                $parentProject = $projectRepository->find( $parent);
            }
            if (!empty($group)) {
                $structureProject = InstanceProvider::getRepository()->getStructure()->find($group);
            }
            if ($dateExceptionType == 'MB') {
                $dateException = null;
            }
            if (!empty($category)) {
                $category = InstanceProvider::getRepository()->getClassification()->find($category);
            }
            if (!empty($projectStatus)) {
                /** @var Entities\TaskStatus $projectStatus */
                $projectStatus = InstanceProvider::getRepository()->getTaskStatus()->find($projectStatus);
                if($projectStatus->getControl() === Entities\TaskStatus::TASK_STATUS_CONTROL_CONCLUDED && Self::getProductProject($project)) {
                    $this->hasChildrenProjectOfTypeDevOps($project->getId());
                }
            }

            \Util\InstanceProvider::getLog()->setLog('muitos finds');
            $project = $this->updateProject($project, $name, $beginDate, $endDate, $status, $structureProject, $parentProject,
                $dateExceptionType, $dateException, $description, $projectType, $category, $projectStatus,
                $productVersionDevId, $isLtsReplication
            );
            \Util\InstanceProvider::getLog()->setLog('updateProject');
            $managerResources = array();
            $resourceRepository = InstanceProvider::getRepository()->getResource();
            if(is_array($managers)){
                foreach ($managers as $managerId) {
                    $managerResources[] = $resourceRepository->find($managerId);
                }
            }else{
                $managerResources[] = $resourceRepository->find($managers);
            }
            $resourceProjectService = InstanceProvider::getService()->getResourceProject();
            \Util\InstanceProvider::getLog()->setLog('find managers');
            $resourceProjectService->setManagersProject($managerResources, $project,$mergeOldManager,$mergeNewManager);
            \Util\InstanceProvider::getLog()->setLog('setManagersProject');
            $sponsorResource = $resourceRepository->find($sponsor);
            $resourceProjectService->updateProjectPartnerSponsorOrManager($sponsorResource, $project,
                ResourceProject::SPONSOR, $mergeSponsor, $recursiveSponsor);
            \Util\InstanceProvider::getLog()->setLog('updateProjectPartnerSponsorOrManager');
            $project->setDuration(InstanceProvider::getService()->getCalendarTakCalendar()->setDuration($project));
            \Util\InstanceProvider::getLog()->setLog('setDuration');
            $this->entityManager->flush();
            \Util\InstanceProvider::getLog()->setLog('flush');
            if($changeProjectStatus){
                $this->setParentStatusProject($project);
            }
            \Util\InstanceProvider::getLog()->setLog('setParentStatusProject');
            $this->setChildrenStatusProject($project);
            \Util\InstanceProvider::getLog()->setLog('setChildrenStatusProject');
            $this->setModelProjectRelationships($project);
            \Util\InstanceProvider::getLog()->setLog('setModelProjectRelationships');
            $connection->commit();
            \Util\InstanceProvider::getLog()->setLog('commit');
            if($setsDevOps){
                $this->setManagersChildrenDevOps($project,$managerResources,$mergeOldManager,$mergeNewManager);
                $this->setsDevOps($project);
            }
            \Util\InstanceProvider::getLog()->setLog('setsDevOps2');
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
        return $project;
    }

    public function hasChildrenProjectOfTypeDevOps($projectId)
    {
        $projects = InstanceProvider::getRepository()->getProject()->findChildrenProjectOfDevOpsType($projectId);
        if(!empty($projects)){
            throw new \Exception(Messages::ERR_EDIT_PROJECT_STATUS_CONCLUDED);
        }
        return true;
    }

    public function setChildrenStatusProject(Entities\Project $project,$setModelRestricted = false)
    {
        if ($project->getProjectStatus()
            && in_array($project->getProjectStatus()->getControl(), self::arrayStatusChildrenReplicate())
        ) {
            /** @var Entities\Project[] $projectChildren */
            $projectChildren = InstanceProvider::getRepository()->getProject()->getProjectChildrenById($project->getId());
            foreach ($projectChildren as $projectChild) {
                if ($projectChild->getProjectStatus()
                    && !in_array($projectChild->getProjectStatus()->getControl(), self::arrayStatusNotModify())
                ) {
                    if ($project->getProjectStatus()->getControl() == self::PROJECT_CANCELED_STATUS
                        || $setModelRestricted
                        || !$this->validateRelationshipModelRestricted($projectChild)
                    ) {
                        $projectChild->setProjectStatus($project->getProjectStatus());
                        $this->entityManager->persist($projectChild);
                        $this->setsDevOps($projectChild);
                        $this->setModelProjectRelationships($projectChild);
                    }
                }
            }
            $this->entityManager->flush();
        }
    }

    public function setManagersChildrenDevOps(Entities\Project $project,$resources,$oldManagerIds,$newManagerIds){

        $this->environment->setDOCurrentUserIdByCurrentPMUser();
        $isRemoveOldManagers = !empty($oldManagerIds);
        $isAddNewManagers = !empty($newManagerIds);
        $projectChildrenIds = DevOpsUtilities::getAllProjectLessIdInArray($project->getId());
        $oldDOManagers = array();
        if($isRemoveOldManagers){
            $oldManagerResources = InstanceProvider::getRepository()->getResource()->findBy(array('id' => $oldManagerIds));
            $oldDOManagers = Utilities::getZhUserByResource($oldManagerResources);
        }
        $zhUsers = array();
        if($isAddNewManagers){
            $zhUsers = Utilities::getZhUserByResource($resources);
        }
        $projectRepository = InstanceProvider::getRepository()->getProject();
        foreach ($projectChildrenIds as $projectChildrenId) {
            /** @var Entities\Project $projectChildren */
            if($project->getId() != $projectChildrenId){
                $projectChildren = $projectRepository->find($projectChildrenId);
                if ($isRemoveOldManagers) {
                    $this->removeManagerDevOps($projectChildren, $oldDOManagers);
                }

                if ($isAddNewManagers) {
                    $this->createManagerDevOps($projectChildren, $zhUsers);
                }
            }

        }
        $this->entityManager->flush();
    }

    public function setModelProjectRelationships(Entities\Project $modelProject)
    {
        list($controlsNotCreated,$taskInDeveloping,$taskForMerge) =
            $this->validateIsPossibleChangeForInHomologation($modelProject);

        $this->validateConcluded($modelProject);

        list($controlsNotCreated, $taskInDeveloping, $taskForMerge) = $this->setsModelProject($modelProject,
            $controlsNotCreated, $taskInDeveloping, $taskForMerge);

        $this->validateHomolagation($modelProject, $controlsNotCreated, $taskInDeveloping, $taskForMerge);
    }

    public function validateIsPossibleChangeForInHomologation(Entities\Project $project,&$controlsNotCreated = array(),
                                                              &$taskInDeveloping = array(),&$taskForMerge = array()){
        if($project->getProjectStatus()
            && $project->getProjectStatus()->getControl() == Entities\TaskStatus::TASK_STATUS_CONTROL_IN_HOMOLOGATION) {

            $allControlForHomologation = Classification::getArrayControlForHomologation();
            $arrayClassification = Classification::getArrayAssociativeByControl();
            $classifications = InstanceProvider::getRepository()->getClassification()->findBy(array('control' => $allControlForHomologation));
            $classificationControls = Utilities::getArrayByArrayObjectWithGet($classifications, array('getControl'));
            foreach ($allControlForHomologation as $control) {
                if (!in_array($control, $classificationControls)) {
                    $controlsNotCreated = array($arrayClassification[$control]);
                };
            }
        }
        return array($controlsNotCreated,$taskInDeveloping,$taskForMerge);
    }

    protected function setStatusAllTasksProject(Entities\Project $project, $status)
    {
        /**
         * @var Entities\Project[] $allSubProject
         * */
        if($status != 'A'){
            $allSubProject = InstanceProvider::getRepository()->getProject()->findBy(array("parent" => $project));
            if (!empty($allSubProject)) {
                foreach ($allSubProject as $subProject) {
                    if ($subProject->getStatus() != 'D' ) {
                        $subProject->setStatus($status);
                    }
                    $this->setStatusAllTasksProject($subProject, $status);
                }
            }
        }
    }

    public function getProjectsByManager($projectExclusionId = null, $projectExclusionStatus = array())
    {
        /** @var Repositories\Project $projectRepository */
        $projectRepository = InstanceProvider::getRepository()->getProject();
        $projects = $projectRepository->findProjectsByManager($this->environment->getCurrentOrganizationId(), $this->environment->getCurrentUserId(), $projectExclusionId, $projectExclusionStatus, $this->projectTypeIds);
        $projects = $this->environment->fixParentsForTreeGrid($projects);
        return $projects;
    }


    public static function getMnemonicsByString($tasks)
    {
        $mnemonics = '';
        $mnemonicArray = array();
        $cont = 0;
        $mnemonicValidate = array();
        foreach($tasks as $task) {
            $string = $task['NAME'];
            if (is_string($string)) {
                $cont++;

                if($cont > 0){
                    $isMnemonic = strpos($string, "[");
                    if($isMnemonic !== false){
                        $string = str_replace(']', ']&', $string);
                        $mnemonics = explode('&', $string);
                        array_pop($mnemonics);
                        foreach($mnemonics as $mnemonic){
                            $mnemonic = explode('[', $mnemonic);
                            if(!empty($mnemonic[1])){
                                $mnemonic = explode(']', $mnemonic[1]);
                            }
                            if(!empty($mnemonic[0])){
                                $mnemonicInfo["NAME"] = $mnemonic[0];
                                $mnemonicInfo["NEWNAME"] = '[' .$mnemonic[0] . ']';
                                if(!in_array($mnemonicInfo["NAME"], $mnemonicValidate)) {
                                 array_push($mnemonicArray, $mnemonicInfo);
                                 array_push($mnemonicValidate, $mnemonicInfo["NAME"]);
                                }
                            }
                        }
                    }

                }

            }
        }
        return $mnemonicArray;
    }

    public function findProjectsMnemonics($valueId, $isCopyTask)
    {
        $projectRepository = InstanceProvider::getRepository()->getProject();
        if($isCopyTask){
            $names = $projectRepository->findTasksNamesFromTaskId($valueId);
        }else{
            $tasks = $projectRepository->findAllProjectsByModel($valueId);
            $names = array();
            $projectIds = array();
            $valueId = "";
            foreach($tasks as $task){
                array_push($names, $task);
                if($valueId){
                    $valueId = $valueId . ',' . $task["ID"];
                }else {
                    $valueId = $task["ID"];
                }
            }
            $tasks = $projectRepository->findTasksNamesFromProject($valueId);
            foreach($tasks as $task){
                array_push($names, $task);
            }
        }
        $allMnemonicString = self::getMnemonicsByString($names);
        return $allMnemonicString;
    }

    public function validateMoveTask(Entities\Task $task, Entities\Project $targetProject)
    {
        $distinctType = $task->getProject()->getProjectType()->getId() != $targetProject->getProjectType()->getId();
        $relatedTasks = InstanceProvider::getRepository()->getTaskRelationship()->findRelatedTaskIds($task->getId());
        if (!empty($relatedTasks) && $distinctType) {
            throw new\Exception('A tarefa "' . $task->getName() . Messages::ERR_RELATED_IN_MOVE_TASK);

        }
        $changedFiles = InstanceProvider::getRepository()->getChangedFiles()->findBy(array('task' => $task));
        if (!empty($changedFiles) && $distinctType) {
            throw new\Exception('A tarefa "' . $task->getName() . Messages::ERR_CHANGE_FILE_IN_MOVE_TASK);

        }
    }
    public function validateRelationshipPrecursor($task,$targetTask){
        return !!InstanceProvider::getRepository()->getTaskRelationship()->findOneBy(array('task' => array($task,$targetTask),'taskRelationship' => array($task,$targetTask), 'type' => TaskPrecursor::PRECURSOR_TYPE));
    }

    /**
     * @param $targetProjectId
     * @param $targetTaskId
     * @param $tasksIds
     * @param $basic
     * @param bool $addDifferentResources
     * @param null $differentResourcesIds
     * @param null $replanningReasonId
     * @param null $originProjectId
     * @param null $taskIdsNotMoved
     * @throws \Doctrine\DBAL\ConnectionException
     * @throws \Exception
     */
    public function moveTask($targetProjectId, $targetTaskId, $tasksIds, $basic, $addDifferentResources = false,
                             $differentResourcesIds = null, $replanningReasonId = null, $originProjectId = null, $taskIdsNotMoved = null)
    {
        $taskRepository = InstanceProvider::getRepository()->getTask();
        /** @var null|Entities\Task $targetTask */
        $targetTask = $targetTaskId ? $taskRepository->find($targetTaskId) : null;
        /** @var Entities\Project $targetProject */
        $targetProject = InstanceProvider::getRepository()->getProject()->find($targetProjectId);
        /** @var Entities\Project $originalProject */
        $originalProject = InstanceProvider::getRepository()->getProject()->find($originProjectId);

        $this->compareTasksProductVersionWithProject($targetProject, $tasksIds);

        if ($targetTask) {
            $dateBegin = $targetTask->getEstimatedBeginDate();
            $targetTask->setIsParent(self::IS_PARENT_TRUE);
        } else {
            $dateBegin = $targetProject->getBeginDate();
        }

        $connection = $this->entityManager->getConnection();

        try {
            $connection->beginTransaction();

            //foreach necessary because $taskIds = Request\Row;
            $realTaskIds = $tasksIds;
            $tasksIds = array();
            foreach($realTaskIds as $taskId) {
                $tasksIds[] = $taskId;
            }

            if($this->validateEqualProjects($tasksIds, $targetProjectId)) {
                $tasksIds = $this->addGeneratedTaskIds($tasksIds);
            }
            $taskIdsForBaselineRegister = $tasksIds;

            if ($addDifferentResources) {
                $this->allocateResourcesInProject($differentResourcesIds, $targetProject);
            }
            $attachmentRepository = InstanceProvider::getRepository()->getAttachment();
            foreach ($tasksIds as $taskId) {
                /** @var Entities\Task $task */
                $task = $taskRepository->find($taskId);
                $this->validateMoveTask($task, $targetProject);
                $task->setEstimatedBeginDate($dateBegin);
                $task->setEstimatedEndDate($dateBegin);
                if($task->getProject()->getId() != $targetProjectId || $this->validateRelationshipPrecursor($task,$targetTask)){
                    $this->removePredecessorTasks($task);
                    $this->removeSuccessorTasks($task);
                }
                if (!$addDifferentResources)  {
                    $this->removeAllocatedResourcesThatNotProject($task, $targetProject);
                }
                $task->setProject($targetProject);
                $task->setParent($targetTask);

                $task->setEstimatedEndDate($task->getEstimatedBeginDate());
                $this->updateRepositoriesTaskChangedFiles($targetProject, $task);
                $this->entityManager->persist($task);
                $attachments = $attachmentRepository->findBy(array('task' => $task));
                try {
                    InstanceProvider::getService()->getAttachment()->attachmentsMoveOrCopy($basic, $task, $attachments, true);
                } catch (\Exception $e) { }
                $this->entityManager->flush();
                $this->moveChildrenTasks($task, $targetProject);

                if($replanningReasonId){
                    InstanceProvider::getService()
                                    ->getTaskReplanningReason()
                                    ->createTaskReplanningReason($replanningReasonId, $taskId, $originProjectId, $targetProject);
                }
                $this->applyOrderToIssues($task);
            }
            if(!empty($taskIdsNotMoved)){
                foreach($taskIdsNotMoved as $taskId) {
                    $taskEntity = $taskRepository->find($taskId);
                    $taskEntity->setParent(null);
                }
            }

            $filterInTasks =  "(" . implode(',', $taskIdsForBaselineRegister) . ")";
            $taskIds = $taskRepository->getChildrenTasks($filterInTasks);

            $taskIdsForBaselineRegister = DevOpsUtilities::getArrayFromArrayAssociative($taskIds);

            if($originalProject->getId() != $targetProject->getId() && !$this->isProjectInPlanning($originalProject)) {
                InstanceProvider::getService()->getTask()->setBaselineTaskMoved($taskIdsForBaselineRegister, $originProjectId);
            }

            if(!$this->isProjectInPlanning($targetProject)){

                foreach ($taskIdsForBaselineRegister as $taskId){
                    $taskService = InstanceProvider::getService()->getTask();
                    $baselineProjectIds = $taskRepository->getBaselineProjectIdsFromTaskId($taskId);

                    $isExistingProject = false;
                    foreach ($baselineProjectIds as $project){
                        if($targetProject->getId() == $project['PROJECT']){
                            $isExistingProject = true;
                        }
                    }

                    if ($isExistingProject){
                        /** @var Entities\Project|null $baselineProject */
                        $baselineProject = $baselineProject = $this->getBaselineProjectRelatedOrNull($targetProject->getId());
                        $task = $taskRepository->getBaselineTask($taskId, $baselineProject->getId());
                        if(!empty($task)){
                            /** @var Entities\Task $baselineTask */
                            $baselineTask = $taskRepository->find($task['ID']);
                            $baselineTask->setBaselineOccurrence(null);
                            $baselineTask->setBaselineOccurrenceAt(null);
                            $this->entityManager->persist($baselineTask);
                        }
                    } else {
                        /** @var Entities\Project|null $baselineProject */
                        $baselineProject = $this->getBaselineProjectRelatedOrNull($targetProject->getId());

                        if($baselineProject){
                            /** @var Entities\Task $task */
                            $task = $taskRepository->find($taskId);

                            $baselineTask = $taskService->createBaselineTaskByTaskCreated($task, $baselineProject);
                            $taskService->createBaselineTypeTaskRelationship($task, $baselineTask);
                        }
                    }
                }
            }


            $this->entityManager->flush();
            $connection->commit();
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
    }

    /**
     * @param $projectId
     * @return object|null
     */
    public function getBaselineProjectRelatedOrNull($projectId)
    {
        $taskRelationshipFilter = array('project' => $projectId, 'type' => Entities\ProjectRelationship::TYPE_TIME_LINE);
        $projectRelationship = InstanceProvider::getRepository()->getProjectRelationship()->findOneBy($taskRelationshipFilter);
        if($projectRelationship){
            /** @var Entities\Project $baselineProject */
            return InstanceProvider::getRepository()->getProject()->find($projectRelationship->getProjectRelationship());
        } else {
            return null;
        }
    }

    public function isProjectInPlanning(Entities\Project $project)
    {
        return $project->getProjectStatus()->getControl() && $project->getProjectStatus()->getControl() == Entities\TaskStatus::TASK_STATUS_CONTROL_PLANNING;
    }

    public function compareTasksProductVersionWithProject($project, $taskIds)
    {
        $tasks = [];
        $projectProductVersionId = $project->getProductVersion() ? $project->getProductVersion()->getId() : null;
        foreach ($taskIds as $taskId) {
            $task = InstanceProvider::getRepository()->getTask()->find($taskId);
            $taskProductVersionDevId = $task->getProductVersionDev() ? $task->getProductVersionDev()->getId() : null;
            if($projectProductVersionId != $taskProductVersionDevId){
                $tasks[] = $task;
            }
        }

        if(!empty($tasks)){
            $tasksNameMsg = '';
            foreach ($tasks as $task) {
                $tasksNameMsg .= '<li>' . $task->getName() . '</li>';
            }
            throw new \Exception(Messages::ERR_MOVE_TASK_DIFFERENT_VERSION . '' . $tasksNameMsg);
        }
    }

    public function applyOrderToIssues($task){
       InstanceProvider::getService()->getTask()->applyOrderToIssuesByTask($task);
    }

    /**
     * @param $infoTasks
     * @param $selectedIssues
     * @param $basicDirectoryUpload
     * @param $isCreatedByIssue
     * @param $classificationId
     * @param $productVersionDevId
     * @param $projectProductIssueId
     * @return array
     * @throws \Doctrine\DBAL\ConnectionException
     */
    public function createTasksByIssues($infoTasks, $selectedIssues, $basicDirectoryUpload, $isCreatedByIssue, $classificationId, $productVersionDevId, $projectProductIssueId)
    {
        $connection = $this->entityManager->getConnection();
        $connection->beginTransaction();
         try {
            $shouldAddResourcesInProject = false;
            $resourcesIdsToAllocateInProject = null;
            $shouldAddResourcesInTask = true;
            $createdTaskIds = array();
            foreach ($infoTasks as &$infoTask) {
                $targetProjectId = $infoTask['projectId'];
                $resources = $infoTask['resourceIds'];
                $targetParentTask = $infoTask['parentId'];
                $taskFeaturesId = $infoTask['taskFeaturesId'];
                $timePlanned = $infoTask['timePlannedId'];
                $description = $infoTask['description'];
                
                $infoTask['taskIds'] = $this->copyTask(
                    $targetProjectId, $targetParentTask, $selectedIssues, $basicDirectoryUpload,
                    $isCreatedByIssue, $shouldAddResourcesInProject, $resourcesIdsToAllocateInProject,
                    $shouldAddResourcesInTask, $resources,$classificationId,$taskFeaturesId,$timePlanned,
                    $productVersionDevId, $projectProductIssueId, null, $description
                );
                $createdTaskIds = array_merge($createdTaskIds, $infoTask['taskIds']);
            }

            $taskRelationshipRepository = InstanceProvider::getRepository()->getTaskRelationship();
            foreach ($createdTaskIds as $TaskId) {
                $taskRepository = InstanceProvider::getRepository()->getTask();
                $task = $taskRepository->find($TaskId);
                $taskRelationshipId = $taskRelationshipRepository->findLastEndDateTask($task->getId(), $task->getProject()->getId(), $this->environment->getCurrentUserId());
                if(!empty($taskRelationshipId)){
                    $taskRelationship = $taskRepository->find($taskRelationshipId['ID']);
                    $taskRelationshipFactory = InstanceProvider::getFactory()->getTaskRelationship();
                    
                    $relationship = $taskRelationshipFactory->createTaskRelationship($task, $taskRelationship, Entities\TaskRelationship::PRECURSOR_TYPE, Entities\TaskRelationship::END_START_TYPE);
                    
                    $this->entityManager->persist($relationship);
                    $this->entityManager->flush();
                }

            }


            $tasks = InstanceProvider::getRepository()->getTask()->getTasksWithSameResource($createdTaskIds);
            $taskService = InstanceProvider::getService()->getTask();
            foreach ($tasks as $task) {
                $taskIds = DevOpsUtilities::setSelectMultiInQuery($task['TASKS']);
                if (count($taskIds) > 1 && !$projectProductIssueId) {
                    $taskService->createTaskRelationshipRelatedTask($taskIds);
                }
            }

            InstanceProvider::getService()->getTask()->createBaselineTasksByTaskIds($createdTaskIds);

            $connection->commit();
            return $createdTaskIds;
        }
        catch(\Exception $e){
            $connection->rollBack();
            throw $e;
        }
    }

    public function createTasksByTask($infoTasks, $primeTaskId, $basicDirectoryUpload)
    {
        $connection = $this->entityManager->getConnection();
        $connection->beginTransaction();
        try {
            $taskRelationshipRepository = InstanceProvider::getRepository()->getTaskRelationship();
            $taskResultIds = $taskRelationshipRepository->findRelatedTaskIds($primeTaskId);
            if(!empty($taskResultIds)){
                $primeTaskId = $taskResultIds['MAIN_TASK'];
            }
            $projectResultIds = $taskRelationshipRepository->findRelatedProjectIds($primeTaskId);
            $taskRepository = InstanceProvider::getRepository()->getTask();
            /** @var Entities\Task $primeTask */
            $primeTask = $taskRepository->find($primeTaskId);
            $projectIds = array();
            if(!empty($projectResultIds)) {
                $projectIds = Utilities::setSelectMultiInQuery($projectResultIds['RELATED_PROJECTS']);
                $projectIds[] = $primeTask->getProject()->getId();
                $projectIds[] = $projectResultIds['MAIN_PROJECT'];
            }
            $shouldAddResourcesInProject = false;
            $resourcesIdsToAllocateInProject = null;
            $shouldAddResourcesInTask = true;
            $createdTaskIds = array();
            foreach ($infoTasks as &$infoTask) {
                $targetProjectId = $infoTask['projectId'];
                $resources = $infoTask['resourceIds'];
                $targetParentTask = $infoTask['parentId'];
                if(!in_array($targetProjectId,$projectIds)){
                    $infoTask['taskIds'] = $this->copyTask(
                        $targetProjectId, $targetParentTask, [$primeTaskId], $basicDirectoryUpload,
                        false, $shouldAddResourcesInProject, $resourcesIdsToAllocateInProject,
                        $shouldAddResourcesInTask, $resources
                    );
                    $createdTaskIds = array_merge($createdTaskIds, $infoTask['taskIds']);
                }
            }
            $createdTaskIdsForBaseLinetask = $createdTaskIds;
            $createdTaskIds[] = $primeTaskId;
            $tasks = $taskRepository->getTasksWithSameResource($createdTaskIds);
            $taskService = InstanceProvider::getService()->getTask();
            foreach ($tasks as $task) {
                $taskIds = DevOpsUtilities::setSelectMultiInQuery($task['TASKS']);
                if (count($taskIds) > 1) {
                    $taskService->createTaskRelationshipRelatedTask($taskIds);
                    foreach ($taskIds as $taskId) {
                        if($taskId != $primeTaskId){
                            $issue = $taskRepository->findIssuesRelatedToTask($primeTaskId, $this->environment->getCurrentOrganizationId(), '');
                            if($issue){
                                $taskService->createTaskRelationship($taskId, [$issue[0]['ID']], TaskPrecursor::RELATIONSHIP_ISSUE_TYPE);
                            }
                        }
                    }
                }
            }
            InstanceProvider::getService()->getTask()->createBaselineTasksByTaskIds($createdTaskIdsForBaseLinetask);
            $connection->commit();
            return $createdTaskIds;
        }
        catch(\Exception $e){
            $connection->rollBack();
            throw $e;
        }
    }

    /**
     * Returns a Task Entity by its id.
     *
     * @param $id
     * @return \Zeedhi\DevOps\Model\Projectmanagement\Entities\Task
     */
    protected function getTaskEntityById($id){
        return $this->entityManager->getRepository(Utilities::ENTITY_DIR.'Task')->find($id);
    }

    /**
     * Copies all task attachments.
     * @param Entities\Task $from
     * @param Entities\Task $to
     */
    protected function copyTaskAttachments(Entities\Task $from, Entities\Task $to, $basicUploadPath)
    {
        $attachments = $this->entityManager->getRepository(Utilities::ENTITY_DIR.'Attachment')->findBy(array('task' => $from));
        try {
            InstanceProvider::getService()->getAttachment()->attachmentsMoveOrCopy($basicUploadPath, $to, $attachments, false);
        } catch(\Exception $e) {}
    }

    public  function copyProjectAttachments(Entities\Project $from, Entities\Project $to)
    {
        $attachments = $this->entityManager->getRepository(Utilities::ENTITY_DIR.'Attachment')->findBy(array('project' => $from,'task' => null));
        try {
            InstanceProvider::getService()->getAttachment()->projectAttachmentsMoveOrCopy( $to, $attachments, false);
        } catch(\Exception $e) {}
    }

    public function deleteAllAttachmentsByProject(Entities\Project $project){
        $attachments = InstanceProvider::getRepository()->getAttachment()->findBy(array('project' => $project));
        InstanceProvider::getService()->getAttachment()->deleteAttachmentByEntities($attachments);
    }

    /**
     * Add resources into a given task.
     *
     * @param $resourcesIds
     * @param Entities\Task $targetTask
     * @throws Exception
     */
    public function allocateResourcesInTask($resourcesIds, Entities\Task $targetTask)
    {
        try {
            $allocationType = null;
            $allocationTaskStatus = null;
            $allocationIsFlushed = true;
            $allocationService = InstanceProvider::getService()->getAllocation();
            foreach($resourcesIds as $id){
                $resource   = $this->entityManager->getRepository(Utilities::ENTITY_DIR.'Resource')->find($id);
                $allocation = $allocationService->createAllocation(
                    $targetTask, $resource, $allocationType,
                    $allocationTaskStatus, $allocationIsFlushed
                );
                $this->entityManager->persist($allocation);
            }
        } catch(\Exception $error){
            throw new Exception($error->getMessage());
        }
    }

    /**
     * Add resources into a given project.
     *
     * @param $resourcesIds
     * @param Entities\Project $targetProject
     * @throws Exception
     */
    protected function allocateResourcesInProject($resourcesIds, Entities\Project $targetProject)
    {
        try {
            $resourceProjectFactory = InstanceProvider::getFactory()->getResourceProject();
            $resourceRepository = InstanceProvider::getRepository()->getResource();
            $resourceProjectRepository = InstanceProvider::getRepository()->getResourceProject();
            foreach ($resourcesIds as $resourceId) {
                $resourceProject = $resourceProjectRepository->findOneBy(array('resourceC' => $resourceId, 'project' => $targetProject));
                if(empty($resourceProject)){
                    $resource = $resourceRepository->find($resourceId);
                    $resourceProjectEntity = $resourceProjectFactory->createResourceProject($resource, $targetProject);
                    $this->entityManager->persist($resourceProjectEntity);
                }
            }
            $this->entityManager->flush();
        } catch (\Exception $error) {
            throw new Exception(Messages::ERR_RESOURCES_NOT_ADDED_EXPLANATION . $error->getMessage());
        }
    }

    /**
     * Copies a task from another already existent. 
     *
     * @param $targetProjectId
     * @param $targetTaskId
     * @param $tasksIds
     * @param $basic
     * @param bool $isCreatedByIssue
     * @param bool $shouldAllocateResourcesInProject
     * @param null $resourcesIdsToAllocateInProject
     * @param bool $shouldAllocateResourcesInTask
     * @param null $resourcesIdsToAllocateInTask
     * @param null $classificationId
     * @return array
     * @throws \Doctrine\DBAL\ConnectionException
     * @throws \Exception
     */
    public function copyTask(
        $targetProjectId, $targetTaskId, $tasksIds, $basic, $isCreatedByIssue = false,
        $shouldAllocateResourcesInProject = false, $resourcesIdsToAllocateInProject = null,
        $shouldAllocateResourcesInTask = false, $resourcesIdsToAllocateInTask = null,
        $classificationId = null, $taskFeaturesId = null, $timePlanned = null, $productVersionDevId = null,
        $projectProductIssueId = null, $useMnemonic = null, $descriptions = null
    )
    {
        $arrayTasks = array();
        foreach ($tasksIds as $tasksId) {
            if (empty($tasksId["TASK_ID"])) {
                $arrayTasks[] = array(
                    "TASK_ID" => $tasksId,
                    "VERSION_ID" => null
                );
            } else {
                $arrayTasks[] = $tasksId;
            }
        }
        $tasksIds = $arrayTasks;
        array_multisort( array_column($tasksIds, "TASK_ID"), SORT_ASC, $tasksIds);
        $totalRating = 0;
        $createdTaskIds = [];
        /** @var null|\Zeedhi\DevOps\Model\Projectmanagement\Entities\Task $targetTask */
        $targetTask = $targetTaskId ? $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'Task')->find($targetTaskId) : null;
        /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Project $targetProject */
        $targetProject = InstanceProvider::getRepository()->getProject()->find($targetProjectId);
        /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Resource[] $resources  */
        $resources = InstanceProvider::getRepository()->getResource()->findBy(array('id' => $resourcesIdsToAllocateInTask));

        if(!$resources){
            $totalRating = 1;
        }
        foreach($resources as $resource){
            if($resource->getRating()){
                $totalRating += $resource->getRating();
            }else{
                $totalRating +=1;
            }
        }
        if ($targetTask) {
            $dateBegin = clone $targetTask->getEstimatedBeginDate();
            $targetTask->setIsParent(self::IS_PARENT_TRUE);
        } else {
            $dateBegin = clone $targetProject->getBeginDate();
        }
        $connection = $this->entityManager->getConnection();
        try {
            $connection->beginTransaction();
            $taskService = InstanceProvider::getService()->getTask();
            foreach ($tasksIds as $taskId) {
                $task = $this->getTaskEntityById($taskId["TASK_ID"]);
                $typeException = $task->getDateExceptionType();
                $taskStatus = $this->getStatusStandard();
                $duration = $task->getDuration();
                $percentCompleted = 0;
                $isParent = $isCreatedByIssue ? 'N' : $task->getIsParent();
                $urlToTest = null;
                $project = $targetProject;
                $description = "";
                if ($project->getProjectStatus()
                    && $project->getProjectStatus()->getControl() == Entities\TaskStatus::TASK_STATUS_CONTROL_IN_HOMOLOGATION
                ) {
                    throw new \Exception(str_replace('%1', $project->getProjectStatus()->getLabel(), Messages::ERR_CLASSIFICATION_IN_HOMOLOGATION), 2);
                }
                $typeException = !empty($typeException) ? $typeException : RestrictionFactory::AS_SOON_AS_POSSIBLE;
                $classification = $task->getClassification();
                if ($classificationId) {
                    $classification = InstanceProvider::getRepository()->getClassification()->find($classificationId);
                }
                if ($classification && $classification->getDivideDuration() == 'N') {
                    $totalRating = 1;
                }
                $duration = !empty($duration) ? $duration : 28800 / $totalRating;
                if ($isCreatedByIssue) {
                    if ($timePlanned && $timePlanned != null) {
                        $duration = $timePlanned;
                    } else {
                        $duration = 0;
                    }
                }

                $taskFeatures = null;
                if ($taskFeaturesId) {
                    $taskFeaturesRepository = InstanceProvider::getRepository()->getTaskFeatures();
                    $taskFeatures = $taskFeaturesRepository->find($taskFeaturesId);
                }
                $taskName = $task->getName();
                if ($useMnemonic) {
                    foreach ($useMnemonic as $mnemonic) {
                        if (strpos($taskName, $mnemonic["NAME"]) !== false) {
                            $taskName = str_replace(("[" . $mnemonic["NAME"] . "]"), $mnemonic["NEWNAME"], $taskName);
                        }
                    }
                }

                if ($descriptions) {
                    $description = empty($descriptions['description_' . $taskId["TASK_ID"]]) ? null : $descriptions['description_' . $taskId["TASK_ID"]];
                }else {
                    $description = $task->getDescription();
                }
                $copiedTask = $taskService->createTask(
                    $targetProject, $task->getLatency(), $taskName, $dateBegin,
                    $dateBegin, $task->getStatus(), $description, $percentCompleted, $duration,
                    $targetTask, $taskStatus, $typeException, $task->getDateException(), $isParent, null,
                    $classification, $taskFeatures, null, $timePlanned, $taskId["VERSION_ID"]);
                    
                $createdTaskIds[] = $copiedTask->getId();
                if(!$timePlanned){
                    $timePlanned = $task->getTimePlanned();
                    $timePlanned = empty($timePlanned) ? $copiedTask->getDuration() * $totalRating : $task->getTimePlanned();
                }
                $copiedTask->setTimePlanned($timePlanned);
                $copiedTask->setProgress($task->getProgress());
                $this->copyChildrenTasks($task, $targetProject, $copiedTask, false, $useMnemonic);
                $this->copyTaskAttachments($task, $copiedTask, $basic);

                if(!empty($resourcesIdsToAllocateInTask)){
                    $this->allocateResourcesInTask($resourcesIdsToAllocateInTask, $copiedTask);
                }
                
                if ($isCreatedByIssue) {
                    $this->createIssueTaskRelationship($task, $copiedTask);
                    
                    $project = $copiedTask->getProject();
                    $projectType = InstanceProvider::getRepository()->getProjectType()->find($project->getProjectType());
                    if($projectType->getExternaltype() == 'E' && $productVersionDevId){
                        $projectProductId = InstanceProvider::getService()->getProject()->getProductProject($project)->getId();
                        if($projectProductId == $projectProductIssueId) {
                            $productVersion = InstanceProvider::getRepository()->getProductVersion()->find($productVersionDevId);
                            $project->setProductVersion($productVersion);
                            $this->entityManager->persist($project);
                      
                            $task->setProductVersionDev($productVersion);
                            
                        } else {
                            $productVersion = $project->getProductVersion();
                        }
                        $copiedTask->setProductVersionDev($productVersion);
                        
                        $this->entityManager->persist($copiedTask);
                    } else {
                        $copiedTask->setProductVersionDev($copiedTask->getProject()->getProductVersion());
                    }
                }
            }

            if ($shouldAllocateResourcesInProject) {
                $this->allocateResourcesInProject($resourcesIdsToAllocateInProject, $targetProject);
            }

            $this->entityManager->flush();
            $task = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'Task')->findOneBy(array('project' => $targetProject));
            
            $this->entityManager->persist($task);
            $this->entityManager->flush();
            $connection->commit();
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
        return $createdTaskIds;
    }

    /**
     * Creates a relationship between an issue and a task.
     *
     * @param $task
     * @param $taskPrecursor
     */
    private function createIssueTaskRelationship($task, $taskPrecursor){
        $typePrecursor = null;
        $isFlushNeeded = false;
        InstanceProvider::getService()->getTaskPrecursor()->createTaskPrecursor(
            $task, $taskPrecursor, TaskPrecursor::RELATIONSHIP_ISSUE_TYPE, $typePrecursor, $isFlushNeeded
        );
    }

    public function getStatusStandard()
    {
        /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskProperties $taskProperties */
        $taskProperties = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskProperties')->findOneBy(array('organization' => $this->environment->getCurrentOrganizationId()));
        return $taskProperties->getInitialTaskStatus();
    }

    protected function removePredecessorTasks($task)
    {
        /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskRelationship[] $predecessorTasks */
        $predecessorTasks = InstanceProvider::getRepository()->getTaskRelationship()->findBy(array('task' => $task, 'type' => TaskPrecursor::PRECURSOR_TYPE));
        foreach ($predecessorTasks as $taskPredecessor) {
            $this->entityManager->remove($taskPredecessor);
            $this->entityManager->flush();
        }
    }

    protected function removeSuccessorTasks($task)
    {
        $successorTasks = InstanceProvider::getRepository()->getTaskRelationship()->findBy(array('taskRelationship' => $task, 'type' => TaskPrecursor::PRECURSOR_TYPE));
        foreach ($successorTasks as $successorTask) {
            $this->entityManager->remove($successorTask);
            $this->entityManager->flush();
        }
    }


    protected function removeAllocatedResources($task)
    {
        try {
            $resourcesAllocated = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\Allocation')->findBy(array('task' => $task));
            foreach ($resourcesAllocated as $resource) {
                $this->entityManager->remove($resource);
                $this->entityManager->flush();
            }
        } catch (\Exception $e) {
            throw new Exception('Não é possível mover tarefa(s) já apropriadas.');
        }
    }

    /************************************************************
     * Add resources in a project
     * @param $tasksIds
     * @param $projectId
     ************************************************************/
    protected function addAllocatedResourcesThatNotProject($resourcesIds, Entities\Project $targetProject)
    {
        try {
            $resourceProjectFactory = InstanceProvider::getFactory()->getResourceProject();
            foreach ($resourcesIds as $id) {
                $resource = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'Resource')->find($id);
                $resourceProjectEntity = $resourceProjectFactory->createResourceProject($resource, $targetProject);
                $this->entityManager->persist($resourceProjectEntity);
            }
            $this->entityManager->flush();
        } catch (\Exception $error) {
            throw new Exception(Messages::ERR_RESOURCES_NOT_ADDED_EXPLANATION . $error->getMessage());
        }
    }

    protected function removeAllocatedResourcesThatNotProject(\Zeedhi\DevOps\Model\Projectmanagement\Entities\Task $task, \Zeedhi\DevOps\Model\Projectmanagement\Entities\Project $targetProject)
    {
        try {
            /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Allocation[] $resourcesAllocated */
            $resourcesAllocated = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\Allocation')->findBy(array('task' => $task));
            /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\ResourceProject[] $resourcesProject */
            $resourcesProject = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\ResourceProject')->findBy(array('project' => $targetProject->getId()));
            $arrayIdResources = array();
            foreach ($resourcesProject as $resourceProject) {
                array_push($arrayIdResources, $resourceProject->getResourceC()->getId());
            }
            foreach ($resourcesAllocated as $allocate) {
                if (!in_array($allocate->getResource()->getId(), $arrayIdResources)) {
                    $this->entityManager->remove($allocate);
                }
            }
            $this->entityManager->flush();
        } catch (\Exception $e) {
            throw new Exception('Não é possível mover tarefa(s) já apropriadas.' . $e->getMessage());
        }
    }

    /**
     * @param \Zeedhi\DevOps\Model\Projectmanagement\Entities\Task $task
     * @param $targetProject
     * @throws Exception
     * @throws \Doctrine\ORM\OptimisticLockException
     */
    private function moveChildrenTasks(\Zeedhi\DevOps\Model\Projectmanagement\Entities\Task $task, $targetProject)
    {
        $childrenTasks = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\Task')->findBy(array('parent' => $task));
        /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Task $childrenTask */
        foreach ($childrenTasks as $childrenTask) {
            if($childrenTask->getTaskStatus()->getId() != $this->environment->getTaskStatusIsConcluded()->getId()){
                $childrenTask->setProject($targetProject);
                $childrenTask->setEstimatedBeginDate($task->getEstimatedBeginDate());
                $this->removePredecessorTasks($childrenTask);
                $this->removeSuccessorTasks($childrenTask);
                $this->removeAllocatedResourcesThatNotProject($childrenTask, $targetProject);
                $this->entityManager->persist($childrenTask);

                $this->moveChildrenTasks($childrenTask, $targetProject);
                $this->entityManager->flush();
            }
        }
    }

    /**
     * @param Entities\Task $task
     * @param $targetProject
     * @param Entities\Task $copyTask
     * @param bool $standard
     * @param null $mnemonicName
     * @param array $allTask
     * @return Entities\Task[]
     * @throws \Doctrine\ORM\OptimisticLockException
     * @throws \Exception
     */
    private function copyChildrenTasks( Entities\Task $task, $targetProject, Entities\Task $copyTask ,$standard = true,
                                        $mnemonicName = null, &$allTask = array() )
    {

        $childrenTasks = InstanceProvider::getRepository()->getTask()->findBy(array('parent' => $task), array('id' => 'ASC'));
        $taskService = InstanceProvider::getService()->getTask();
        /** @var Entities\Task $childrenTask */
        foreach ($childrenTasks as $childrenTask) {
            if ($standard) {
                $isParent = Entities\Task::NOT_IS_PARENT;
                $urlToTest = null;
            } else {
                $isParent = $childrenTask->getIsParent();
                $urlToTest = $childrenTask->getUrlToTest();
            }
            $name = $this->validateMnemonic($mnemonicName, $childrenTask->getName());
            $copyChildrenTask = $taskService->createTask($targetProject, $childrenTask->getLatency(),
                $name, $copyTask->getEstimatedBeginDate(),
                $childrenTask->getEstimatedEndDate(), $childrenTask->getStatus(), $childrenTask->getDescription(),
                $childrenTask->getPercentCompleted(), $childrenTask->getDuration(), $copyTask,
                $this->getStatusStandard(), $task->getDateExceptionType(), $task->getDateException(), $isParent, $urlToTest,
                $task->getClassification());
            $copyChildrenTask->setResponse($childrenTask->getResponse());
            $copyChildrenTask->setTimePlanned($childrenTask->getTimePlanned());
            $copyChildrenTask->setProgress($childrenTask->getProgress());
            $allTask[] = array('copy'=>$copyChildrenTask,'origin' =>$childrenTask );
            $this->entityManager->flush();
            $this->copyChildrenTasks($childrenTask, $targetProject, $copyChildrenTask, $standard, $mnemonicName, $allTask);
        }
        return $allTask;
    }

    public function validateMnemonic($mnemonicName, $name){
        if($mnemonicName){
            foreach($mnemonicName as $mnemonic) {
                if (strpos($name, $mnemonic["NAME"]) !== false) {
                    $newMnemonicName =  $mnemonic["NEWNAME"];
                    $name = str_replace(("[" . $mnemonic["NAME"] . "]"), $newMnemonicName, $name);
                    if($name == "") {
                        throw new \Exception('O Mnemônico ' . "[" . $mnemonic["NAME"] . "]" . ' não pode ser nulo');
                    }
                }
            }
        }
        return $name;
    }


    public function findProjectByOrganization()
    {
        $projectsAll = InstanceProvider::getRepository()->getProject()
            ->findProjectByOrganization($this->environment->getCurrentOrganizationId());
        foreach ($projectsAll as &$project) {
            $project['PERCENT_COMPLETED'] = $project['PERCENT_COMPLETED'] . '%';
        }
        return $projectsAll;
    }

    public function findProjectLog($projectId, $filters = [])
    {
        $filters = $this->environment->createQueryFiltered($filters, 'P');
        $findProjectLog = InstanceProvider::getRepository()->getProject()->findProjectLog($projectId, $filters);
        return $findProjectLog;
    }

    public function findProjectByUser()
    {

        $userId = $this->environment->getCurrentUserId();
        $projectsAll = InstanceProvider::getRepository()->getProject()->findProjectByOrganization($userId);
        $projectsAllIDs = array();
        foreach ($projectsAll as $project) {
            array_push($projectsAllIDs, $project['ID']);
        }
        foreach ($projectsAll as &$project) {
            if (!in_array($project['PARENT'], $projectsAllIDs)) {
                $project['PARENT'] = null;
            }
        }
        return $projectsAll;
    }

    public function validateExternalType($projectTypeId, array $externalTypes){
        $validation = false;
        if($projectTypeId){
            /** @var Entities\ProjectType $projectType */
            $projectType = InstanceProvider::getRepository()->getAttachment()->find($projectTypeId);
            $validation = in_array($projectType->getExternaltype(),$externalTypes);
        }
        return $validation;
    }

    public function findProjectParent($filterColumn)
    {
        $filterString = Utilities::createQueryFiltered($filterColumn);
        $projectsAll = InstanceProvider::getRepository()->getProject()->findProjectParent($this->environment->getCurrentOrganizationId(), $filterString);
        $projectsAll = $this->environment->fixParentsForTreeGrid($projectsAll);
        return $projectsAll;
    }

    public function findManagerByProject($projectId)
    {
        $projectsAll = InstanceProvider::getRepository()->getProject()->findManagerByProject($projectId);
        return $projectsAll;
    }

    public function findSponsorByProject($projectId)
    {
        $projectsAll = InstanceProvider::getRepository()->getProject()->findSponsorByProject($projectId);
        return $projectsAll;
    }

    public function newProject($name, $parent, $status, $duration, $endDate, $beginDate, $group, $managerIds, $sponsor,
                               $dateExceptionType, $dateException, $projectType = null, $category = null,
                               $description = null, $projectStatus = null,$projectModelId = null, $productVersionDevId = null, $mnemonicName = null, $isLtsReplication = 'Y'){
        $name = $this->validateMnemonic($mnemonicName, $name);
        if($projectModelId) {
            $project = $this->createNewProjectByModel($name, $parent, $status, $duration, $endDate, $beginDate, $group, $managerIds, $sponsor,
                $dateExceptionType, $dateException, $projectType, $category, $description, $projectStatus, $projectModelId, $productVersionDevId, $mnemonicName, $isLtsReplication);
        } else {
            $project = $this->createNewProject($name, $parent, $status, $duration, $endDate, $beginDate, $group, $managerIds, $sponsor,
                $dateExceptionType, $dateException, $projectType, $category, $description, $projectStatus, $productVersionDevId, $isLtsReplication);
        }
        return $project;
    }

    public function createNewProjectByModel($name, $parentId, $status, $duration, $endDate, $beginDate, $group, $managerIds, $sponsorId,
                                            $dateExceptionType, $dateException, $projectType, $categoryId, $description, $projectStatusId,
                                            $projectModelId, $productVersionDevId, $mnemonicName = null, $isLtsReplication = 'Y')
    {

        $connection = $this->entityManager->getConnection();
        $connection->beginTransaction();
        $project = null;
        try {
            $projectRepository = InstanceProvider::getRepository()->getProject();
            $projectModel = $projectRepository->find($projectModelId);
            /** @var Entities\Project $projectModelFirst */
            $projectModelFirst = $projectRepository->findOneBy(array('parent' => $projectModel));
            $managerResources = array();
            $resourceRepository = InstanceProvider::getRepository()->getResource();
            foreach ($managerIds as $managerId) {
                $managerResources[] = $resourceRepository->find($managerId);
            }

            $sponsor = $resourceRepository->find($sponsorId);
            $project = $this->createNewProject($name, $parentId, $status, $duration, $endDate, $beginDate, $group, $managerIds, $sponsorId,
                $dateExceptionType, $dateException, $projectType, $categoryId, $description, $projectStatusId, $productVersionDevId, $isLtsReplication);

            $projectRelationshipService = InstanceProvider::getService()->getProjectRelationship();
            $projectRelationshipService->createProjectRelationship($project, $projectModelFirst,
                Entities\ProjectRelationship::TYPE_MODEL);
            $projectRelationshipService->createProjectRelationship($project, $projectModel,
                Entities\ProjectRelationship::TYPE_MODEL_GENERATION);
            list($tagBegin, $tagEnd) = $this->getTagNameModel($projectModelFirst, $project);
            $this->copyAllTask($projectModelFirst,$project, $mnemonicName);
            $this->copyProjectChildrenModel($projectModelFirst, $project, $managerResources, $sponsor, $projectModel,
                $tagBegin, $tagEnd, $mnemonicName);
            $this->entityManager->flush();
            $connection->commit();
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
        return $project;

    }

    public function cloneResourceAndRepositoriesProject(Entities\Project $project)
    {
        $productProject = self::getProductProject($project->parent());
        $projectParent = $productProject ?: $project->getParent();
        if ($projectParent) {
            $parentRepositories = InstanceProvider::getRepository()->getRepository()->findBy(array('project' => $projectParent));
            $parentResourceProject = InstanceProvider::getRepository()->getResourceProject()->findBy(array('project' => $projectParent));
            $this->cloneResourceProject($project, $parentResourceProject);
            $this->cloneProjectRepositories($project, $parentRepositories, $parentResourceProject);
        }
    }

    public function copyProjectChildrenModel($project,$projectTarget,$managerResources,$sponsor,$projectModel,
                                             $tagBegin, $tagEnd, $mnemonicName)
    {
        $projectChildren = InstanceProvider::getRepository()->getProject()->findBy(array('parent' => $project));

        foreach ($projectChildren as $projectChild) {
            $copyProjectChild = $this->copyProjectModel($projectChild, $projectTarget, $managerResources,$sponsor, $mnemonicName);
            $copyProjectChild->setName($tagBegin . $copyProjectChild->getName() . $tagEnd);
            $projectRelationshipService = InstanceProvider::getService()->getProjectRelationship();
            $projectRelationshipService->createProjectRelationship($copyProjectChild, $projectChild,
                Entities\ProjectRelationship::TYPE_MODEL);
            $projectRelationshipService->createProjectRelationship($copyProjectChild, $projectModel,
                Entities\ProjectRelationship::TYPE_MODEL_GENERATION);
            $this->setsDevOps($copyProjectChild);
            $this->cloneResourceAndRepositoriesProject($copyProjectChild);
            $this->copyProjectChildrenModel($projectChild, $copyProjectChild, $managerResources, $sponsor, $projectModel,
                $tagBegin, $tagEnd, $mnemonicName);
        }
    }

    public function copyProjectModel(Entities\Project $projectChild,Entities\Project $projectParent,$managerResources,$sponsor, $mnemonicName){
        if ($projectChild->getProjectStatus()->getControl() == "D") {
            $projectChild->setProjectStatus(Utilities::getProjectStatusPlanning());
        }
        $productVersion = null;
        if($projectParent->getProductVersion()){
            $productVersion = $projectParent->getProductVersion();
        }

        $name = $this->validateMnemonic($mnemonicName, $projectChild->getName());

        $copyProjectChild = $this->createProject($name, $projectParent, $projectChild->getStatus(),
            $projectChild->getDuration(), $projectChild->getEndDate(), $projectChild->getBeginDate(), null, null,
            $projectParent->getStructure(), $projectChild->getDateExceptionType(), $projectChild->getDateException(),
            $projectChild->getProjectType(), $projectChild->getClassification(), $projectChild->getDescription(),
            $projectChild->getProjectStatus(), $productVersion, $projectChild->getIsLtsReplication());

        $resourceProjectService = InstanceProvider::getService()->getResourceProject();
        foreach ($managerResources as $managerResource) {
            $resourceProjectService->createResourceProjectFast($managerResource, $copyProjectChild, ResourceProject::MANAGER);
        }
        $resourceProjectService->createResourceProjectFast($sponsor, $copyProjectChild, ResourceProject::SPONSOR);
        $this->copyAllTask($projectChild,$copyProjectChild, $mnemonicName);
        $this->entityManager->flush();
        return $copyProjectChild;
    }

    public function copyAllTask(Entities\Project $project,Entities\Project $targetProject, $mnemonicName = null){
        /** @var Entities\Task[] $allTaskFathers */
        $allTaskFathers = InstanceProvider::getRepository()->getTask()->findBy(array('project' => $project,'parent' => null), array('id' => 'ASC'));
        $allTask = array();
        $taskService = InstanceProvider::getService()->getTask();
        foreach ($allTaskFathers as $task) {
            $name = $this->validateMnemonic($mnemonicName, $task->getName());

            $copiedTask = $taskService->createTask(
                $targetProject, $task->getLatency(), $name, $task->getEstimatedBeginDate(),
                $task->getEstimatedBeginDate(), $task->getStatus(), $task->getDescription(), $task->getPercentCompleted(),
                $task->getDuration(), null, $task->getTaskStatus(), $task->getDateExceptionType(), $task->getDateException(), $task->getIsParent(),
                $task->getUrlToTest(), $task->getClassification());
            $copiedTask->setTimePlanned($task->getTimePlanned());
            $copiedTask->setResponse($task->getResponse());
            $copiedTask->setProgress($task->getProgress());
            $allTask[] = array('copy'=>$copiedTask,'origin' =>$task );
            $this->entityManager->flush();
            $allTask = $this->copyChildrenTasks($task, $targetProject, $copiedTask, false, $mnemonicName, $allTask);
        }
        $this->copyTaskPrecursors($allTask);
        $this->entityManager->flush();
    }

    public function createNewProject($name, $parent, $status, $duration, $endDate, $beginDate, $group, $managerIds, $sponsor,
                                     $dateExceptionType, $dateException, $projectType = null, $category = null, $description = null,
                                     $projectStatus = null, $productVersionDevId = null, $isLtsReplication = 'Y')
    {

        $connection = $this->entityManager->getConnection();
        $connection->beginTransaction();
        $calendarRepository = InstanceProvider::getRepository()->getCalendar();
        $project = null;
        try {
            $parentRepositories = null;
            $parentResourceProject = null;
            $productVersion = null;
            if (!empty($parent)) {
                $parent = $this->entityManager->find(Utilities::ENTITY_DIR . 'Project', $parent);
                $parentRepositories = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'Repository')->findBy(array('project' => $parent->getId()));
                $parentResourceProject = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'ResourceProject')->findBy(array('project' => $parent->getId()));
            }
            $group = $this->entityManager->find(Utilities::ENTITY_DIR . 'Structure', $group);
            if(!is_array($managerIds)){
                $managerIds = array($managerIds);
            }
            /** @var Entities\Resource $manager */
            $managerResources = array();
            $resourceRepository = InstanceProvider::getRepository()->getResource();
            foreach ($managerIds as  $managerId){
                $managerResources[] = $resourceRepository->find($managerId);
            }
            if (!InstanceProvider::getService()->getResource()->isManagers($managerResources)) {
                throw new \Exception('O recurso indicado como gerente não pode ser um gerente de projetos');
            }


            $sponsor = $this->entityManager->find(Utilities::ENTITY_DIR . 'Resource', $sponsor);
            if (!empty($category)) {
                $category = $this->entityManager->find(Utilities::ENTITY_DIR . 'Classification', $category);
            }
            if ($projectType) {
                $projectType = $this->entityManager->find(Utilities::ENTITY_DIR . 'ProjectType', $projectType);
            }
            if ($projectStatus) {
                $projectStatus = $this->entityManager->find(Utilities::ENTITY_DIR . 'TaskStatus', $projectStatus);
            }
            if($productVersionDevId){
                $productVersion = InstanceProvider::getRepository()->getProductVersion()->find($productVersionDevId);
            } else {
                $productVersion = null;
            }

            $project = $this->createProject($name, $parent, $status, 3600 * 8, $endDate, $beginDate, $managerResources, $sponsor
                , $group, $dateExceptionType, $dateException, $projectType, $category, $description, $projectStatus, $productVersion, $isLtsReplication);

            if($parent && $parent->getCalendar()){
                $calendar = $calendarRepository->find($parent->getCalendar());
                $this->saveEditCalendar(null, $calendar->getCalendarDefault()->getId(), $project->getId(), false);
            }

            $productProject = self::getProductProject($project->parent());
            if($productProject){
                $parentRepositories = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'Repository')->findBy(array('project' => $productProject));
                $parentResourceProject = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'ResourceProject')->findBy(array('project' => $productProject));
            }
            $this->cloneProjectRepositories($project, $parentRepositories, $parentResourceProject);
            $this->cloneResourceProject($project, $parentResourceProject);

            $resourceProjectService = InstanceProvider::getService()->getResourceProject();
            foreach ($managerResources as $managerResource) {
                $resourceProjectService->createResourceProject($managerResource, $project->getId(), ResourceProject::MANAGER);
            }

            $resourceProjectService->createResourceProject($sponsor, $project->getId(), ResourceProject::SPONSOR);
            $this->entityManager->flush();
            $connection->commit();
            $this->setsDevOps($project);
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
        return $project;
    }

    /**
     * @param Entities\Project $project
     * @param Entities\Repository[] $parentRepositories
     * @param Entities\ResourceProject[] $resourcesProject
     * @return Entities\Repository[]
     */
    public function cloneProjectRepositories(Entities\Project $project, $parentRepositories, $resourcesProject)
    {
        $repositories = array();
        if (!empty($parentRepositories)) {
            $repositoryService = InstanceProvider::getService()->getRepository();
            foreach ($parentRepositories as $parentRepository) {
                $repository = $repositoryService->saveProjectRepository($project, $parentRepository->getStatus(),
                    $parentRepository->getDescription(), $parentRepository->getName(), $parentRepository->getUrl(),
                    $parentRepository->getPath(), $parentRepository->getSeqno());
                array_push($repositories, $repository);
                $this->cloneResourceDirectories($resourcesProject, $repository, $parentRepository);
            }
        }
        return $repositories;
    }

    /**
     * @param Entities\Project $project
     * @param Entities\ResourceProject[] $parentResourcesProject
     * @param bool $withType
     * @return array
     */
    public function cloneResourceProject($project, $parentResourcesProject,$withType = false)
    {
        $resources = array();
        if (!empty($parentResourcesProject)) {
            $resourceProjectService = InstanceProvider::getService()->getResourceProject();
            $resourceProjectRepository = InstanceProvider::getRepository()->getResourceProject();
            foreach ($parentResourcesProject as $parentResource) {
                $type = $withType?$parentResource->getType():resourceProject::WORK;
                $resource = $resourceProjectRepository->findOneBy(array('resourceC' => $parentResource->getResourceC(), 'project' => $project));
                if(!$resource){
                    $resource = $resourceProjectService->createResourceProject($parentResource->getResourceC()->getId(), $project, $type);
                }
                array_push($resources, $resource);
            }
        }
        return $resources;
    }


    /**
     * @param Entities\ResourceProject[] $resourcesProject
     * @param Entities\Repository $repository
     * @param Entities\Repository $parentRepository
     */
    public function cloneResourceDirectories($resourcesProject, $repository, $parentRepository)
    {
        if (!empty($resourcesProject)) {
            $directoryRepository = InstanceProvider::getRepository()->getDirectory();
            foreach ($resourcesProject as $resourceProject) {
                /** @var Entities\ResourceProject $resourceProject */
                $directory = $directoryRepository->findOneBy(array('resource' => $resourceProject->getResourceC(),
                    'repository' => $parentRepository));
                /** @var Entities\Directory $directory */
                if ($directory) {
                    InstanceProvider::getService()->getResourceProject()->saveDirectory($repository->getId(), $directory->getLocalDirectory(),
                        $resourceProject->getResourceC()->getId(), null, $directory->getLinuxDirectory());
                }
            }
        }
    }

    public function saveCalendarNewProject(\Zeedhi\DevOps\Model\Projectmanagement\Entities\Project $project, $name, $status, $beginWeek, $endWeek, $hour, $beginHour, $endHour,
                                           $weeklyHours, $monthlyDays, $businessPartner, $asset, $entity)
    {

        $connection = $this->entityManager->getConnection();
        $connection->beginTransaction();
        try {
            $calendarService = InstanceProvider::getService()->getCalendar();
            $calendar = $calendarService->saveEditCalendar($project->getId(), $name, $status, $beginWeek, $endWeek, $hour,
                $beginHour, $endHour, $weeklyHours, $monthlyDays, $businessPartner, $asset, null, $entity);
            $project->getCalendar();
            $this->entityManager->persist($project);
            $this->entityManager->flush();
            $connection->commit();

        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
        return $calendar;
    }

    public function getDateException($taskId, $type)
    {
        /**
         * @var \Zeedhi\DevOps\Model\Projectmanagement\Repositories\DateExceptionType $repositoryDateException
         */
        $repositoryDateException = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\DateExceptionType');
        /**
         * @var \Zeedhi\DevOps\Model\Projectmanagement\Repositories\Task $repositoryTask
         */
        $repositoryTask = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\Task');
        $allParent = $repositoryTask->getParentTask($taskId);

        if (empty($allParent) && $type == 'T') {
            $allException = $repositoryDateException->getAllDateException();
        } else {
            $allException = $repositoryDateException->getAllDateExceptionForTaskUp();
        }
        return $allException;
    }

    public function saveEditException($calendarEntity, $commitmentType, $name, \DateTime $beginDate, \DateTime $endDate,
                                      $status, $unit, $timeBreak, $beginTimeBreak, $endTimeBreak, $commitmentId)
    {

        InstanceProvider::getService()->getCalendar()->saveEditException($calendarEntity, $commitmentType, $name, $beginDate, $endDate,
            $status, $unit, $timeBreak, $beginTimeBreak, $endTimeBreak, $commitmentId);

    }

    public function findProducts($projectId = 0, $filterName = false, $filterOperator = false, $filterValue= false)
    {

        $organization = $this->environment->getCurrentOrganizationId();
        $products = InstanceProvider::getRepository()->getProject()->findProducts($organization, $projectId, $filterName, $filterOperator, $filterValue);
        $notModelProducts = DevOpsUtilities::removeModels($products,$organization);
        return $notModelProducts;
    }

    public function findProductsExceptConcludedOrCanceled($conditions = '', $page, $pageSize)
    {
        $filterString = Utilities::createQueryFiltered($conditions);
        $organization = $this->environment->getCurrentOrganizationId();
        $limit = $page * $pageSize;
        $offset = ($page - 1) * $pageSize;
        $products = InstanceProvider::getRepository()->getProject()->findProductsExceptConcludedOrCanceled($organization, $filterString, $limit, $offset);
        $notModelProducts = DevOpsUtilities::removeModels($products,$organization);
        return $notModelProducts;
    }

    public function findAllProductsBudget($conditions, $page, $pageSize)
    {
        $filterString = Utilities::createQueryFiltered($conditions);
        $organization = $this->environment->getCurrentOrganizationId();
        $limit = $page * $pageSize;
        $offset = ($page - 1) * $pageSize;
        $products = InstanceProvider::getRepository()->getProject()->findAllProductsBudget($organization, $filterString, $limit, $offset);
        $notModelProducts = DevOpsUtilities::removeModels($products,$organization);
        return $notModelProducts;
    }



    public function findMyProducts($projectId = 0, $conditions, $structures)
    {
        $organization = $this->environment->getCurrentOrganizationId();
        $projectStatusId = DevopsUtilities::getValueInFilterDataSourceByName($conditions, 'PROJECT_STATUS_ID');
        if(!$projectStatusId) {
            DevopsUtilities::getValueInFilterDataSourceByName($conditions, 'PROJECT_STATUS_ID', true);
        }
        $filters = Utilities::createQueryFiltered($conditions, 'P');
        $structures = DevOpsUtilities::getStringForArray($structures);
        $structures = $structures ? $this->getQueryFilterStructureId($structures) : $structures;
        $products = InstanceProvider::getRepository()->getProject()->findMyProducts($organization, $projectId,$this->environment->getCurrentUserId(), $filters, $structures);
        return DevOpsUtilities::removeModels($products,$organization);
    }

    private function getQueryFilterStructureId($structureIdsString)
    {
        return "AND P.STRUCTURE IN (SELECT S.ID
                                      FROM STRUCTURE S
                                     WHERE STATUS = 'A'
                                START WITH S.ID IN ($structureIdsString)
                          CONNECT BY PRIOR S.ID = S.PARENT)";
    }

    public function findAllProducts()
    {
        return InstanceProvider::getRepository()->getProject()->findAllProducts($this->environment->getCurrentOrganizationId());
    }

    public function saveNewBoard($name, $manager, $structure, $parent, $description)
    {
        $connection = $this->entityManager->getConnection();
        try {
            $connection->beginTransaction();
            $board = InstanceProvider::getRepository()->getProject()
                ->findOneBy(array("organization" => $this->environment->getCurrentOrganizationId(), "name" => $name));

            if (empty($board)) {
                $manager = $this->entityManager->find('\Zeedhi\DevOps\Model\Projectmanagement\Entities\Resource', $manager);

                $structure = $this->entityManager->find('\Zeedhi\DevOps\Model\Projectmanagement\Entities\Structure', $structure);

                if (!empty($parent)) {
                    $parent = $this->entityManager->find('\Zeedhi\DevOps\Model\Projectmanagement\Entities\Project', $parent);
                }

                $board = InstanceProvider::getFactory()->getProject()->createProject($name, $parent, self::ACTIVE, null, null, null, $structure, null, null, $description,
                    self::IS_BOARD);
                $this->entityManager->persist($board);
                $boardManager = InstanceProvider::getService()->getTaskPrecursor()->createResourceProject($manager, $board, self::PROJECT_MANAGER_TYPE);
                $this->entityManager->persist($boardManager);
                $this->entityManager->flush();
                $connection->commit();
            } else {
                $connection->rollBack();
            }
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
    }

    public function saveBoard($board, $name, $manager, $structure, $parent, $description)
    {
        $connection = $this->entityManager->getConnection();
        try {
            $connection->beginTransaction();
            $board = InstanceProvider::getRepository()->getProject()->find($board);
            $manager = $this->entityManager->find('\Zeedhi\DevOps\Model\Projectmanagement\Entities\Resource', $manager);
            $structure = $this->entityManager->find('\Zeedhi\DevOps\Model\Projectmanagement\Entities\Structure', $structure);
            if (!empty($parent)) {
                $parent = $this->entityManager->find('\Zeedhi\DevOps\Model\Projectmanagement\Entities\Project', $parent);
            }
            $board = InstanceProvider::getFactory()->getProject()->updateProject($board, $name, null, null, self::ACTIVE, $structure, $parent, null, null, null, $description);
            $this->entityManager->persist($board);
            $boardManager = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\ResourceProject')
                ->findOneBy(array('project' => $board->getId(), 'type' => self::PROJECT_MANAGER_TYPE));
            InstanceProvider::getService()->getTaskPrecursor()->updateProjectPartner($boardManager, $manager, $board, self::PROJECT_MANAGER_TYPE);
            $this->entityManager->flush();
            $connection->commit();
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
    }

    public function getProjectById($projectId)
    {
        return $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\Project')->findOneBy(array("id" => $projectId));
    }

    public function getTopMostManagerStructure($userId, $structures)
    {
        $permissionRepository = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\Permission');
        foreach ($structures as $structure) {
            $permission = $permissionRepository->findBy(array('structure' => $structure['ID'], 'resource' => $userId));
            if (!empty($permission) && empty($structure['PARENT'])) {
                return $structure;
            }
        }
    }

    public function getBoardData($board)
    {
        return InstanceProvider::getRepository()->getProject()->selectBoard($board);
    }


    public function isMergeRequestProject($projectId)
    {
        /** @var Entities\Project $project */
        $project = InstanceProvider::getRepository()->getProject()->find($projectId);
        $projectType = $project->getProjectType();
        $isMergeRequestProject = $projectType->getUsemergerequest();
        return $isMergeRequestProject;
    }

    public function findProjectCategories($projectType = null, $filterLabel = '')
    {
        $filterLabelString = DevOpsUtilities::createQueryFiltered($filterLabel);
        $projectRepository = InstanceProvider::getRepository()->getProject();
        return $projectRepository->findProjectCategories($this->environment->getCurrentOrganizationId(), $projectType, $filterLabelString);
    }


    public function findProjectTypesFilter($filterName, $filterOperator, $filterValue)
    {
        $projectRepository = InstanceProvider::getRepository()->getProject();
        return $projectRepository->findProjectTypes($this->environment->getCurrentOrganizationId(), $filterName, $filterOperator, $filterValue);
    }

    public function findProjectTypesToIssue($filter)
    {
        $filterString = DevOpsUtilities::createQueryFiltered($filter);
        return InstanceProvider::getRepository()->getProjectType()->findProjectTypes($this->environment->getCurrentOrganizationId(),$filterString);
    }

    /**
     * Return date begin max and date end min
     * @param $projectId
     * @return array
     */
    protected function getDatesRestrictionsMinAndMax($projectId)
    {
        $dateBeginMax = null;
        $dateEndMin = null;
        try {
            /** @var \Zeedhi\DevOps\Model\Projectmanagement\EntitiesInterface\Calendar\TaskCalendarInterface $projectEntity */
            $projectEntity = InstanceProvider::getRepository()->getProject()->find($projectId);
            if ($projectEntity->hasChildren() && $projectEntity->usesDataControl()) {
                $children = $projectEntity->children();
                $taskCalendar = InstanceProvider::getService()->getCalendarTakCalendar();
                foreach ($children as $child) {
                    /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Task $child */
                    if ((!$child->isContainer() || $child->hasChildren()) && $projectEntity->usesDataControl()) {
                        list($dateBeginMax, $dateEndMin) = $taskCalendar->getRestriction($child)->getDatesMinAndMax($dateBeginMax, $dateEndMin);
                    }
                }
            } else {
                $dateBeginMax = null;
                $dateEndMin = null;
            }

        } catch (\Exception $e) {
            $dateBeginMax = null;
            $dateEndMin = null;
        }
        return array($dateBeginMax, $dateEndMin);
    }

    /*******************************************************************
     * Gets a list of all project status stored in database.
     *
     * @param $filterName
     * @param $filterOperator
     * @param $filterValue
     * @return array
     *****************************************************************
     */
    public function findProjectStatus($filterName = false, $filterOperator = false, $filterValue = false, $conditions = null)
    {
        $filter = Utilities::createQueryFiltered($conditions);
        $organizationId = $this->environment->getCurrentOrganizationId();
        return InstanceProvider::getRepository()->getProject()->findProjectStatusByOrganization($organizationId, $filterName, $filterOperator, $filterValue, $filter);
    }

    public function findProjectStatusInEditProject($filters)
    {
        $filterString = Utilities::createQueryFiltered($filters);
        $organizationId = $this->environment->getCurrentOrganizationId();
        $projectStatus = InstanceProvider::getRepository()->getProject()->findProjectStatusInEditProject($filterString,$organizationId);
        return $projectStatus;
    }

    public function findProjectStatusByControl($concludedStatus)
    {
        $organizationId = $this->environment->getCurrentOrganizationId();
        $projectStatus = InstanceProvider::getRepository()->getProject()->findProjectStatusByControlAndOrganization($concludedStatus, $organizationId);
        return $projectStatus;
    }

    /*******************************************************************
     * Updates an existent project status
     *
     * @param $statusId
     * @param $statusLabel
     * @param string $statusGeneratesLog
     * @param string $statusTaskCanStart
     * @param null $control
     * @throws \Doctrine\ORM\ORMException
     * @throws \Doctrine\ORM\OptimisticLockException
     * @throws \Doctrine\ORM\TransactionRequiredException
     *******************************************************************/
    public function updateProjectStatus($statusId, $statusLabel, $statusGeneratesLog = 'Y', $statusTaskCanStart = 'Y',
                                        $control = null)
    {
        $entityTaskStatus = $this->entityManager->find(\Util\Utilities::ENTITY_DIR . 'TaskStatus', $statusId);
        $entityTaskStatus = InstanceProvider::getFactory()->getProject()->updateProjectStatus($entityTaskStatus, $statusLabel, $statusGeneratesLog,
            $statusTaskCanStart, $control);
        $this->entityManager->persist($entityTaskStatus);
        $this->entityManager->flush();
    }

    /*******************************************************************
     * Creates a new project status and stores it into database.
     *
     * @param $statusLabel
     * @param $statusGeneratesLog
     * @param $statusTaskCanStart
     * @param null $control
     *******************************************************************/
    public function createProjectStatus($statusLabel, $statusGeneratesLog, $statusTaskCanStart, $control = null)
    {
        $entityTaskStatus = InstanceProvider::getFactory()->getProject()->createProjectStatus($statusLabel, $statusGeneratesLog,
            $statusTaskCanStart, $control);
        $this->entityManager->persist($entityTaskStatus);
        $this->entityManager->flush();
        return $entityTaskStatus;
    }

    /*******************************************************************
     * Removes a specified project status from database.
     *
     * @param $statusId
     * @throws \Doctrine\ORM\ORMException
     * @throws \Doctrine\ORM\OptimisticLockException
     * @throws \Doctrine\ORM\TransactionRequiredException
     *******************************************************************/
    public function removeProjectStatus($statusId)
    {
        $entityTaskStatus = $this->entityManager->find(\Util\Utilities::ENTITY_DIR . 'TaskStatus', $statusId);
        $this->entityManager->remove($entityTaskStatus);
        $this->entityManager->flush();
    }

    public function findBasicCalendar($projectId)
    {
        /** @var Entities\Project $project */
        $project = InstanceProvider::getRepository()->getProject()->find($projectId);
        if ($project && $project->getCalendar() && $project->getCalendar()->getCalendarDefault()) {
            $calendars = InstanceProvider::getService()->getCalendar()->repositoryCalendar->getCalendar($project->getCalendar()->getId());
            foreach ($calendars as &$calendar) {
                $calendar['PROJECT'] = $projectId;
            }
        } else if($project){
            $structure = $project->getStructure();
            $name = $this->environment->getNextEntityWithCalendar($structure)->getName();
            $calendars = array(array(
                'PARENT_CALENDAR_TEXT' => Messages::ALE_CALENDAR_IS_NOT_OWN . $name,
                'CALENDAR_DEFAULT' => "",
                'ID' => "",
                'PROJECT' => $projectId));
        }else{
            $calendars = array(array(
                'PARENT_CALENDAR_TEXT' => Messages::ALE_CALENDAR_IS_NOT_OWN ,
                'CALENDAR_DEFAULT' => "",
                'ID' => "",
                'PROJECT' => $projectId));
        }
        return $calendars;
    }

    public function findBasicCalendarResource($projectResourceId)
    {
        /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\ResourceProject $resourceProject */
        $resourceProject = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'ResourceProject')->find($projectResourceId);
        if ($resourceProject && $resourceProject->getCalendar()) {
            $calendars = InstanceProvider::getService()->getCalendar()->repositoryCalendar->getCalendar($resourceProject->getCalendar()->getId());
            foreach ($calendars as &$calendar) {
                $calendar['PROJECT_RESOURCE'] = $projectResourceId;
            }
        } else {
            $calendars = array(array(
                'PARENT_CALENDAR_TEXT' => Messages::ALE_CALENDAR_IS_NOT_OWN . 'do proprio recurso',
                'CALENDAR_DEFAULT' => "",
                'ID' => "",
                'PROJECT_RESOURCE' => $projectResourceId));
        }
        return $calendars;
    }

    public function findExceptionCalendarDatesProject($projectId, $beginDate, $endDate)
    {

        /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Project $project */
        $project = InstanceProvider::getRepository()->getProject()->find($projectId);
        $structure = $project->getStructure();
        return InstanceProvider::getService()->getCalendar()->findExceptionCalendarDatesProject($project, $structure, $beginDate, $endDate);
    }

    public function findExceptionCalendarDatesProjectResource($projectResourceId, $beginDate, $endDate)
    {

        /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\ResourceProject $projectResource */
        $projectResource = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'ResourceProject')->find($projectResourceId);
        return InstanceProvider::getService()->getCalendar()->findExceptionCalendarDatesProjectResource($projectResource, $beginDate, $endDate);
    }

    public function findCalendarExceptionByProject($projectId, $id, $filter = '')
    {

        /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Project $project */
        $project = InstanceProvider::getRepository()->getProject()->find($projectId);
        $structure = $project->getStructure();
        return InstanceProvider::getService()->getCalendar()->findCalendarExceptionByProject($project, $structure, $id, $filter);
    }

    public function findCalendarExceptionByProjectResource($projectResourceId, $id)
    {
        /** @var Entities\ResourceProject $projectResource */
        $projectResource = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'ResourceProject')->find($projectResourceId);
        return InstanceProvider::getService()->getCalendar()->findCalendarExceptionByProjectResource($projectResource, $id);
    }

    public function findCalendarWeeklyDays($projectId, $id, $filter = '')
    {

        /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Project $project */
        $project = InstanceProvider::getRepository()->getProject()->find($projectId);
        $structure = $project->getStructure();
        return InstanceProvider::getService()->getCalendar()->findCalendarWeeklyDaysByProject($project, $structure, $id, $filter);
    }

    public function findCalendarProjectResourceWeeklyDays($projectResourceId, $id)
    {

        /** @var Entities\ResourceProject $projectResource */
        $projectResource = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'ResourceProject')->find($projectResourceId);
        return InstanceProvider::getService()->getCalendar()->findCalendarWeeklyDaysByProjectResource($projectResource, $id);
    }

    public function calculationBasedStart($taskId, $begin, $end, $duration, $parent, $exception, $exceptionDate, $group)
    {

        if ($parent) {
            $parent = InstanceProvider::getRepository()->getProject()->find($parent);
        }
        if ($exceptionDate) {
            $exceptionDateTask = \DateTime::createFromFormat(\Util\Utilities::FORMAT_DATE_DEFAULT, $exceptionDate . ' 00:00:00');
        } else {
            $exceptionDateTask = null;
        }
        if ($taskId) {
            /** @var Entities\Project $task */
            $task = InstanceProvider::getRepository()->getProject()->find($taskId);
            /** @var Entities\Structure $group */
            if ($group) {
                $group = InstanceProvider::getRepository()->getTask()->find($group);
            } else {
                $user = $this->environment->getCurrentUser();
                /** @var Entities\Permission $permission */
                $permission = InstanceProvider::getRepository()->getPermission()->findOneBy(array('resource' => $user));
                $group = $permission->getStructure();
            }
            $task->setDateException($exceptionDateTask);
            $task->setDateExceptionType($exception);
            $task->setStructure($group);
            $task->setEstimatedBeginDate($begin);
            $task->setEstimatedendDate($end);
            $task->setParent($parent);
        } else {
            $group = InstanceProvider::getRepository()->getTask()->find($group);
            $task = InstanceProvider::getFactory()->getProject()->createProject(null, $parent, 'A', 0, $end, $begin, $group, $exception, $exceptionDateTask);
        }

        $taskCalendar = InstanceProvider::getService()->getCalendarTakCalendar();
        list($begin, $end, $exception, $exceptionDate) = $taskCalendar->calculationBasedStart($task, $begin,
            $duration, null, null, $parent, $exception, $exceptionDate);
        $end = $taskCalendar->getEndDateByCalendarInFrontEnd($task,clone $begin,
            $taskCalendar->getDurationInMilliseconds($duration, $task));
        return array($begin, $end, $exception, $exceptionDate);
    }

    public function calculationBasedEnd($taskId, $begin, $end, $duration, $parent, $exception, $exceptionDate, $group)
    {

        if ($parent) {
            /** @var Entities\Project $task */
            $parent = InstanceProvider::getRepository()->getProject()->find($parent);
        }
        if ($exceptionDate) {
            $exceptionDateTask = \DateTime::createFromFormat(\Util\Utilities::FORMAT_DATE_DEFAULT, $exceptionDate . ' 00:00:00');
        } else {
            $exceptionDateTask = null;
        }
        if ($taskId) {
            /** @var Entities\Project $task */
            $task = InstanceProvider::getRepository()->getProject()->find($taskId);

            /** @var Entities\Structure $group */
            $group = InstanceProvider::getRepository()->getTask()->find($group);
            $task->setDateException($exceptionDateTask);
            $task->setDateExceptionType($exception);
            $task->setStructure($group);
            $task->setEstimatedBeginDate($begin);
            $task->setEstimatedendDate($end);
            $task->setParent($parent);
        } else {
            $group = InstanceProvider::getRepository()->getTask()->find($group);
            $task = InstanceProvider::getFactory()->getProject()->createProject(null, $parent, 'A', 0, $end, $begin, $group, $exception, $exceptionDateTask);
        }

        $taskCalendar = InstanceProvider::getService()->getCalendarTakCalendar();
        list($begin, $end, $exception, $exceptionDate) = $taskCalendar->calculationBasedEnd($task, $end, $duration,
            null, null, $parent, $exception, $exceptionDate);
        $begin = $taskCalendar->getBeginDateByCalendarInFrontEnd($task,clone $end,
            $taskCalendar->getDurationInMilliseconds($duration, $task));
        return array($begin, $end, $exception, $exceptionDate);
    }

    public function saveEditCalendar($calendarId, $calendarDefaultId, $projectId, $isApplyChildrenProject = false)
    {
        $connection = $this->entityManager->getConnection();
        try {
            $connection->beginTransaction();
            $calendarRepository = InstanceProvider::getRepository()->getCalendar();
            $projectRepository = InstanceProvider::getRepository()->getProject();
            $project = $projectRepository->find($projectId);
            $nameCalendar = $project->getId() . '_' . $project->getName();
            $calendarService = InstanceProvider::getService()->getCalendar();
            $calendar = $calendarService->saveEditCalendar($calendarId, $nameCalendar, null, null, 'A', Calendar::CALENDAR_RESOURCE);

            if ($calendarDefaultId) {
                $calendarDefault = $calendarService->find($calendarDefaultId);
                $calendar->setCalendarDefault($calendarDefault);
                $project->setCalendar($calendar);
                $this->entityManager->persist($calendar);
                $this->entityManager->persist($project);
                $this->entityManager->flush();
            } else {
                $project->setCalendar(null);
                $calendarService->calendarDelete($calendar);
            }

            $calendarReturn = $calendar;

            if($isApplyChildrenProject){
                $projectsChildren = $projectRepository->getProjectChildrenNotCanceledAndNotConcludedById($projectId);
                if(!empty($projectsChildren)){
                    if($calendarDefaultId){
                        foreach($projectsChildren as $projectChildren){
                            $this->saveEditCalendarChildrenProject($projectChildren, $calendarDefaultId);
                        }
                    } else {
                        foreach($projectsChildren as $projectChildren){
                            if($projectChildren->getCalendar()){
                                $calendar = $calendarRepository->find($projectChildren->getCalendar());
                                if($calendar){
                                    $projectChildren->setCalendar(null);
                                    $calendarService->calendarDelete($calendar);
                                }
                            }
                        }
                    }
                }
            }

            $connection->commit();
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
        return $calendarReturn;
    }

    public function saveEditCalendarChildrenProject(Entities\Project $project, $calendarDefaultId)
    {
        $connection = $this->entityManager->getConnection();
        try {
            $connection->beginTransaction();
            $calendarRepository = InstanceProvider::getRepository()->getCalendar();

            $calendarService = InstanceProvider::getService()->getCalendar();
            if($project->getCalendar()){
                $calendar = $calendarRepository->find($project->getCalendar());
                $project->setCalendar(null);
                $calendarService->calendarDelete($calendar->getId());
            }

            $nameCalendar = $project->getId() . '_' . $project->getName();
            $calendar = $calendarService->saveEditCalendar(null, $nameCalendar, null, null, 'A', Calendar::CALENDAR_RESOURCE);

            $calendarDefault = $calendarService->find($calendarDefaultId);
            $calendar->setCalendarDefault($calendarDefault);
            $project->setCalendar($calendar);
            $this->entityManager->persist($calendar);
            $this->entityManager->persist($project);
            $this->entityManager->flush();

            $connection->commit();
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }

        return $calendar;
    }

    public function saveEditCalendarResource($id, $calendarDefaultId, $projectResourceId)
    {
        $connection = $this->entityManager->getConnection();
        try {
            $connection->beginTransaction();
            /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\ResourceProject $resourceProject */
            $resourceProject = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'ResourceProject')
                ->find($projectResourceId);
            $nameCalendar = $resourceProject->getId() . '_' . $resourceProject->getResourceC()->getName();
            $calendarService = InstanceProvider::getService()->getCalendar();
            $calendar = $calendarService->saveEditCalendar($id, $nameCalendar, null, null, 'A', Calendar::CALENDAR_ALLOCATION);
            if ($calendarDefaultId) {
                $calendarDefault = $calendarService->find($calendarDefaultId);
                $calendar->setCalendarDefault($calendarDefault);
                $resourceProject->setCalendar($calendar);
                $this->entityManager->persist($calendar);
                $this->entityManager->persist($resourceProject);
                $this->entityManager->flush();
            } else {
                $resourceProject->setCalendar(null);
                $calendarService->calendarDelete($calendar);
            }

            $connection->commit();
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
        return $calendar;
    }

    public function beforeSaveWeeklyDays($id, $calendarId, $name, $typeSunday, $dailyHoursSunday, $typeMonday,
                                   $dailyHoursMonday, $typeTuesday, $dailyHoursTuesday, $typeWednesday, $dailyHoursWednesday, $typeThursday,
                                   $dailyHoursThursday, $typeFriday, $dailyHoursFriday, $typeSaturday, $dailyHoursSaturday, $beginDate, $endDate,
                                   $projectId = null, $isApplyChildrenProject = false)
    {
        $calendarDetail = $this->saveWeeklyDays($id, $calendarId, $name, $typeSunday, $dailyHoursSunday, $typeMonday,
                               $dailyHoursMonday, $typeTuesday, $dailyHoursTuesday, $typeWednesday, $dailyHoursWednesday, $typeThursday,
                               $dailyHoursThursday, $typeFriday, $dailyHoursFriday, $typeSaturday, $dailyHoursSaturday, $beginDate, $endDate);

        if($isApplyChildrenProject && !$id){
            $projectRepository  = InstanceProvider::getRepository()->getProject();
            $calendarRepository = InstanceProvider::getRepository()->getCalendar();
            /** @var Entities\Project[] $childrenProjects */
            $childrenProjects = $projectRepository->getProjectChildrenNotCanceledAndNotConcludedById($projectId);

            if(!empty($childrenProjects)){
                $calendarParent = $calendarRepository->find($calendarId);
                $calendarDefaultId = $calendarParent->getCalendarDefault()->getId();

                foreach($childrenProjects as $projectChildren){

                    if(!$projectChildren->getCalendar()){
                        $calendar = $this->saveEditCalendarChildrenProject($projectChildren, $calendarDefaultId);
                    } else {
                        $calendar = $calendarRepository->find($projectChildren->getCalendar());
                    }

                    $this->saveWeeklyDays($id, $calendar->getId(), $name, $typeSunday, $dailyHoursSunday, $typeMonday,
                                          $dailyHoursMonday, $typeTuesday, $dailyHoursTuesday, $typeWednesday, $dailyHoursWednesday,
                                          $typeThursday, $dailyHoursThursday, $typeFriday, $dailyHoursFriday, $typeSaturday,
                                          $dailyHoursSaturday, $beginDate, $endDate);
                }
            }
        }

        return $calendarDetail;
    }

    public function saveWeeklyDays($id, $calendar, $name, $typeSunday, $dailyHoursSunday, $typeMonday,
                   $dailyHoursMonday, $typeTuesday, $dailyHoursTuesday, $typeWednesday, $dailyHoursWednesday, $typeThursday,
                   $dailyHoursThursday, $typeFriday, $dailyHoursFriday, $typeSaturday, $dailyHoursSaturday, $beginDate, $endDate)
    {

        $calendarDetail = InstanceProvider::getService()->getCalendar()->saveWeeklyDays($id, $calendar, $name, $typeSunday, $dailyHoursSunday, $typeMonday,
                                                $dailyHoursMonday, $typeTuesday, $dailyHoursTuesday, $typeWednesday, $dailyHoursWednesday,
                                                $typeThursday, $dailyHoursThursday, $typeFriday, $dailyHoursFriday, $typeSaturday,
                                                $dailyHoursSaturday, $beginDate, $endDate);

        return $calendarDetail;
    }

    /************************************************************
     * Temporally method, because there is another similar in
     * another branch that wasn't accepted yet.
     ************************************************************/
    private function arrayToInQueryStatement($array)
    {
        return "(" . implode(", ", $array) . ")";
    }

    /************************************************************
     * Gets different resources between project and task
     * @param $tasksIds
     * @param $projectId
     * @return array
     ***********************************************************/
    public function getDifferentResourcesBetweenProjectAndTask($tasksIds, $projectId)
    {
        return InstanceProvider::getRepository()->getProject()->getDifferentResourcesBetweenProjectAndTask($this->arrayToInQueryStatement($tasksIds), $projectId);
    }

    /**
     * @param Product $product
     * @param Entities\Project $project
     * @param ZhUser[] $DOManagers
     * @return Product
     */
    protected function saveProduct(Product $product = null, Entities\Project $project,$DOManagers)
    {
        $product = InstanceProvider::getFactory()->getProduct()->saveProduct($product, $project);
        $this->entityManager->persist($product);
        $this->entityManager->flush();
        InstanceProvider::getService()->getZhUserAllocation()->saveManagerProduct($product,$DOManagers);
        return $product;
    }

    /**
     * @param Release $release
     * @param Entities\Project $project
     * @param string $nameRelease
     * @param \Zeedhi\DevOps\Model\Configurationmanager\Entities\Product $product
     * @param ZhUser[] $DOManagers
     * @param $isFixRelease
     * @param null $productServer
     * @return Release
     */
    protected function saveRelease(Release $release = null, Entities\Project $project, $nameRelease, $product, $DOManagers,
                                   $isFixRelease, $productServer = null)
    {
        if($release && $release->getProductServer()) {
            $productServer = $release->getProductServer();
        }
        $release = InstanceProvider::getFactory()->getRelease()->saveRelease($release, $project, $nameRelease, $product, $isFixRelease,
            $productServer);
        $this->entityManager->persist($release);
        $this->entityManager->flush();
        InstanceProvider::getService()->getZhUserAllocation()->saveManagerRelease($release,$DOManagers);
        return $release;
    }

    protected function saveSubRelease(Subrelease $subRelease = null, $release, $project, $name, $DOManagers, $productServer, DevopsEntities\ProductVersion $productVersion = null)
    {
        if($subRelease && $subRelease->getProductServer()) {
            $productServer = $subRelease->getProductServer();
        }

        $subRelease = InstanceProvider::getFactory()->getSubrelease()->saveSubRelease($subRelease, $release, $project, $name,
            $productServer, $productVersion);
        $this->entityManager->persist($subRelease);
        $this->entityManager->flush();
        InstanceProvider::getService()->getZhUserAllocation()->saveManagerSubRelease($subRelease,$DOManagers);
        return $subRelease;
    }

    /**
     * @param Entities\Project $project
     * @param ZhUser[] $DOManagers
     * @return null|object|\Zeedhi\DevOps\Model\Configurationmanager\Entities\Release
     * @throws \Exception
     */
    protected function setDevOpsWhenProjectTypeRelease(Entities\Project $project,array $DOManagers)
    {
        $product = $this->getProductByProject($project);
        $release = null;
        if($product){
            $productServer = $this->getProductServerForRelease($product);
            $release = InstanceProvider::getRepository()->getRelease()->findOneBy(array('project' => $project));
            $clone = false;
            if ($product && !$release) {
                $clone = true;
            }
            $release = $this->saveRelease($release, $project, $project->getName(), $product,
                $DOManagers, self::NOT_IS_FIX_RELEASE, $productServer);
            if ($clone) {
                $this->cloneAllPackage($product, $release);
            }
        }
        return $release;
    }

    protected function cloneAllPackage(Product $product, Release $release){
        /** @var  Package[] $productPackages */
        $productPackages = InstanceProvider::getRepository()->getPackage()->findBy(array('product' => $product));
        $packages = array();

        foreach ($productPackages as $productPackage) {
            $package = clone $productPackage;
            $package->setRelease($release);
            $this->entityManager->persist($package);
            $this->entityManager->flush();
            $this->cloneAllPackageContents($productPackage, $package);

            $packages[] = $package;
        }
        $this->entityManager->flush();
        return $packages;
    }

    /**
     * @param Entities\Project $project
     * @return Product
     * @throws \Exception
     */
    protected function getProductByProject(Entities\Project $project)
    {
        $productProject = self::getProductProject($project);
        $product = InstanceProvider::getRepository()->getProduct()->findOneBy(array('project' => $productProject));
        if (!$product || !$productProject) {
            throw new \Exception(Messages::DEO_ERROR_RELEASE_WITHOUT_PRODUCT,2);
        }
        return $product;
    }

    /**
     * @param ZhUser[] $DOManagers
     */
    protected function setManagersDecOps(array $DOManagers){

    }

    /**
     * @param Entities\Project $project
     * @param ZhUser[] $DOManagers
     * @return null|object|\Zeedhi\DevOps\Model\Configurationmanager\Entities\Product
     */
    protected function setDevOpsWhenProjectTypeProduct(Entities\Project $project, $DOManagers)
    {
        /** @var Product $product */
        $product = InstanceProvider::getRepository()->getProduct()->findOneBy(array('project' => $project));
        $product = $this->saveProduct($product, $project, $DOManagers);
        /** @var Release $release */
        $release = InstanceProvider::getRepository()->getRelease()->findOneBy(array('project' => $project));
        $nameRelease = Messages::DEO_CREATE_RELEASE_NAME_PREFIX . $project->getName();
        $release = $this->saveRelease($release, $project, $nameRelease, $product, $DOManagers, self::IS_FIX_RELEASE);
        $this->entityManager->persist($release);

        $this->setProductInAllProjectRepositories($project, $product);

        $this->entityManager->flush();
        return $product;
    }

    /**
     * @param Entities\Project $project
     * @param ZhUser[] $DOManagers
     * @return null|object | SubRelease
     * @throws \Exception
     */
    protected function setDevOpsWhenProjectTypeSubRelease(Entities\Project $project,array $DOManagers)
    {
        $releaseProject = self::getReleaseProject($project);
        if (!$releaseProject) {
            $this->setDevOpsWhenProjectTypeRelease($project, $DOManagers);
            $releaseProject = $project;
        }
        $release = InstanceProvider::getRepository()->getRelease()->findOneBy(array('project' => $releaseProject));
        $subRelease = InstanceProvider::getRepository()->getSubrelease()->findOneBy(array('project' => $project));
        if($release){
            $product = $this->getProductByProject($project);
            $productServer = $this->getProductServerForRelease($product, 'T');
            $subRelease = $this->saveSubRelease($subRelease, $release, $project, $project->getName(), $DOManagers,
                $productServer);
        }
        return $subRelease;
    }

    /**
     * @param Entities\Project $project
     * @param ZhUser[] $DOManagers
     * @return null|object|\Zeedhi\DevOps\Model\Configurationmanager\Entities\SubRelease
     * @throws \Exception
     */
    protected function setDevOpsWhenProjectTypeError(Entities\Project $project, $DOManagers, $productVersion = null, $nameConcat = '')
    {
        $releaseProject = self::getProductProject($project);
        if (!$releaseProject) {
            throw new \Exception(Messages::DEO_ERROR_RELEASE_WITHOUT_PRODUCT,2);
        }
        $release = InstanceProvider::getRepository()->getRelease()->findOneBy(array('project' => $releaseProject));
        /** @var Subrelease $subRelease */
        $subRelease = InstanceProvider::getRepository()->getSubrelease()->findOneBy(array('project' => $project, 'productVersion' => $productVersion));
        if($release){
            $productServer = $this->getProductServerForRelease($release->getProduct(), 'T');
            $subReleaseName = $project->getName() . $nameConcat;
            $subRelease = $this->saveSubRelease($subRelease, $release, $project, $subReleaseName, $DOManagers,
                $productServer, $productVersion);
        }
        return $subRelease;
    }

    /**
     * @param Entities\Project $project
     * @param $product
     * @return Entities\Repository
     */
    protected function setProductInAllProjectRepositories(Entities\Project $project, $product)
    {
        /** @var Entities\Repository[] $repositories */
        $repositories = InstanceProvider::getRepository()->getRepository()->findBy(array('project' => $project));
        foreach ($repositories as &$repository) {
            $repository->setProduct($product);
            $this->entityManager->persist($repository);
        }
    }

    /**
     * @param $productPackage
     * @param $package
     * @return PackageContents
     */
    protected function cloneAllPackageContents($productPackage, $package)
    {
        $productPackageContents = InstanceProvider::getRepository()->getPackageContents()->findBy(array('package' => $productPackage));
        $packageContents = array();
        /** @var PackageContents[] $productPackageContents */
        foreach ($productPackageContents as $productPackageContent) {
            $packageContent = clone $productPackageContent;
            $packageContent->setPackage($package);
            $this->entityManager->persist($packageContent);
            $packageContents[] = $packageContent;
        }
        return $packageContents;
    }

    protected function addGeneratedTaskIds($taskIds) {

        $realTaskIds = array();
        //foreach necessary because $taskIds = Request\Row;
        foreach($taskIds as $taskId) {
            $realTaskIds[] = $taskId;
        }
        $childIds = [];
        $parentIds = [];
        foreach ($realTaskIds as $taskId) {
            $realTaskIds = $this->getChildGeneratedTasks($taskId, $realTaskIds, $childIds);
            $realTaskIds = $this->getParentGeneratedTasks($taskId, $realTaskIds, $parentIds);
        }
        foreach ($childIds as $childId) {
            $realTaskIds = $this->getParentGeneratedTasks($childId, $realTaskIds, $parentIds);
        }
        foreach ($parentIds as $parentId) {
            $realTaskIds = $this->getChildGeneratedTasks($parentId, $realTaskIds, $childIds);
        }
        $checkAllRelatedTasks = InstanceProvider::getRepository()->getTaskRelationship()->checkAllRelatedTasks($taskIds);
        foreach ($checkAllRelatedTasks as $checkAllRelatedTask) {
            if(!in_array($checkAllRelatedTask['ID'], $realTaskIds)){
                $realTaskIds[] = $checkAllRelatedTask['ID'];
            }
        }
        return $realTaskIds;
    }

    protected function getChildGeneratedTasks($taskId, &$taskIds, &$childIds) {
        /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskRelationship[] $relationships */
        $relationships = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'TaskRelationship')
            ->findBy(array('task' => $taskId));

        foreach($relationships as $relation) {
            $isCallRecursive = !in_array($relation->getTask()->getId(), $taskIds) || !in_array($relation->getTaskRelationship()->getId(), $taskIds);
            if(!in_array($relation->getTaskRelationship()->getId(), $taskIds) &&
                in_array($relation->getType(), self::$DATA_SET_GENERATED_TYPE_RELATIONSHIP_TASK)) {
                $taskIds[] = $relation->getTaskRelationship()->getId();
                $childIds[] = $relation->getTaskRelationship()->getId();
            }

            if(($relation->getType() == Entities\TaskRelationship::MERGE_REQUEST_TYPE ||
                    $relation->getType() == Entities\TaskRelationship::ERROR_MR_TYPE) &&
                $isCallRecursive) {
                return $this->getChildGeneratedTasks($relation->getTaskRelationship()->getId(), $taskIds, $childIds);
            }
        }
        return $taskIds;
    }

    protected function getParentGeneratedTasks($taskId, &$taskIds, &$parentIds) {
        /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskRelationship[] $relationships */
        $relationships = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'TaskRelationship')
            ->findBy(array('taskRelationship' => $taskId));

        foreach($relationships as $relation) {
            $isCallRecursive = !in_array($relation->getTask()->getId(), $taskIds) || !in_array($relation->getTaskRelationship()->getId(), $taskIds);
            if(!in_array($relation->getTask()->getId(), $taskIds) &&
                in_array($relation->getType(), self::$DATA_SET_GENERATED_TYPE_RELATIONSHIP_TASK)) {
                $taskIds[] = $relation->getTask()->getId();
                $parentIds[] = $relation->getTask()->getId();
            }

            if(($relation->getType() == Entities\TaskRelationship::MERGE_REQUEST_TYPE ||
                    $relation->getType() == Entities\TaskRelationship::ERROR_MR_TYPE) &&
                $isCallRecursive) {
                return $this->getParentGeneratedTasks($relation->getTask()->getId(), $taskIds, $parentIds);
            }
        }
        return $taskIds;
    }

    protected function setProjectManagerRecursive($managerId, Entities\Project $meeting)
    {
        /** @var Entities\Resource $managerResource */
        $managerResource = InstanceProvider::getRepository()->getResource()->find($managerId);
        InstanceProvider::getService()->getTaskPrecursor()->updateProjectPartnerSponsorOrManager($managerResource, $meeting,
            ResourceProject::MANAGER, true,true);
    }

    /**
     * @param array $taskIds
     * @param $targetProject
     * @return bool
     */
    protected function validateEqualProjects(Array $taskIds, $targetProject) {
        /** @var Entities\Task $task */
        $task = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'Task')->find($taskIds[0]);
        return $task->getProject()->getId() != $targetProject;
    }

    /**
     * @return array
     */
    protected function getConditionsProjectStatusControlDefault()
    {
        return array(
            "columnName" => 'PROJECT_STATUS_CONTROL',
            "operator" => 'IN',
            "value" => array( Entities\TaskStatus::TASK_STATUS_CONTROL_PLANNING ,
                Entities\TaskStatus::TASK_STATUS_CONTROL_IN_PROGRESS )
        );
    }
    /**
     * @return array
     */
    protected function getConditionsStatusDefault()
    {
        return array(
            "columnName" => 'STATUS',
            "operator" => 'IN',
            "value" => array(Utilities::ACTIVE)
        );
    }

    /**
     * @param array $conditions
     * @return array
     */
    protected function setConditionsMyProjectDefault(array $conditions = array())
    {
        $conditions[] = $this->getConditionsProjectStatusControlDefault();
        $conditions[] = $this->getConditionsStatusDefault();
        return $conditions;
    }

    public function validateRepositoriesMoveTask($tasks, $projectId)
    {
        $taskRepositories = array();
        $projectRepositories = $this->getUrlsProject($projectId);

        $originProjectUrl = null;
        foreach ($tasks as $task) {
            $taskId = $task['ID'];

            if (empty($originProjectUrl)) {
                /** @var Entities\Task $taskEntity */
                $taskEntity = InstanceProvider::getRepository()->getTask()->find($taskId);
                $originProjectUrl = $this->getUrlsProject($taskEntity->getProject()->getId());

            }
            $taskRepositories[$taskId] = array('NAME' => $task['NAME'], 'REPOSITORIES' => array());
            /** @var Entities\ChangedFiles[] $taskChangedFiles */
            $taskChangedFiles = InstanceProvider::getRepository()->getChangedFiles()->findBy(array('task' => $taskId));
            foreach ($taskChangedFiles as $changedFile) {
                $url = $changedFile->getRepository()->getUrl();
                if (in_array($url, $originProjectUrl) &&
                    !isset($taskRepositories[$url]) &&
                    !in_array($url, $projectRepositories)
                ) {
                    $taskRepositories[$taskId]['REPOSITORIES'][$url] = array(
                        'NAME' => $changedFile->getRepository()->getName(),
                        'URL' => $changedFile->getRepository()->getUrl()
                    );
                }

            }
            $taskRepositories[$taskId]['REPOSITORIES'] = array_values($taskRepositories[$taskId]['REPOSITORIES']);
        }

        return array_values($taskRepositories);
    }

    /**
     * @param $allTask
     */
    protected function copyTaskPrecursors($allTask)
    {
        foreach ($allTask as $task) {
            $originTask = $task['origin'];
            $copyTask = $task['copy'];
            $taskRelationships = InstanceProvider::getRepository()->getTaskRelationship()->findBy(array('task' => $originTask,
                'type' => Entities\TaskRelationship::PRECURSOR_TYPE));
            $taskPrecursorService = InstanceProvider::getService()->getTaskPrecursor();
            /** @var Entities\TaskRelationship[] $taskRelationships */
            foreach ($taskRelationships as $taskRelationship) {
                $taskPrecursor = $taskRelationship->getTaskRelationship();
                foreach ($allTask as $task2) {
                    if ($task2['origin']->getId() == $taskPrecursor->getId()) {
                        $taskPrecursorService->createTaskPrecursor($copyTask, $task2['copy'], $taskRelationship->getType(),
                            $taskRelationship->getTypePrecursor());
                        break;
                    }
                }
            }
        }
    }

    protected function removeManagerDevOps(Entities\Project $project,array $oldDOManager)
    {
        if ($project->getProjectType() && $project->getProjectType()->getExternaltype()) {
            $externalType = $project->getProjectType()->getExternaltype();
            $zhUserAllocationService = InstanceProvider::getService()->getZhUserAllocation();
            $releaseRepository = InstanceProvider::getRepository()->getRelease();
            if ($externalType == self::PROJECT_TYPE_PRODUCT) {
                $product = InstanceProvider::getRepository()->getProduct()->findOneBy(array('project' => $project));
                $zhUserAllocationService->removeManagerProduct($product, $oldDOManager);
            } else if ($externalType == self::PROJECT_TYPE_RELEASE) {
                $release= $releaseRepository->findOneBy(array('project' => $project));
                $zhUserAllocationService->removeManagerRelease($release, $oldDOManager);
            } else if ($externalType == self::PROJECT_TYPE_SUB_RELEASE) {
                $subRelease = InstanceProvider::getRepository()->getSubrelease()->findOneBy(array('project' => $project));
                $zhUserAllocationService->removeManagerSubRelease($subRelease, $oldDOManager);
                $release = $releaseRepository->findOneBy(array('project' => $project));
                $zhUserAllocationService->removeManagerRelease($release, $oldDOManager);
            } else if ($externalType == self::PROJECT_TYPE_ERROR) {
                $subRelease = InstanceProvider::getRepository()->getSubrelease()->findOneBy(array('project' => $project));
                $zhUserAllocationService->removeManagerSubRelease($subRelease, $oldDOManager);
            }
        }
    }

    protected function createManagerDevOps(Entities\Project $project,array $zhUsers)
    {
        if ($project->getProjectType() && $project->getProjectType()->getExternaltype()) {
            $externalType = $project->getProjectType()->getExternaltype();
            $zhUserAllocationService = InstanceProvider::getService()->getZhUserAllocation();
            $releaseRepository = InstanceProvider::getRepository()->getRelease();
            if ($externalType == self::PROJECT_TYPE_PRODUCT) {
                $product = InstanceProvider::getRepository()->getProduct()->findOneBy(array('project' => $project));
                $zhUserAllocationService->createManagerProduct($product, $zhUsers);
            } else if ($externalType == self::PROJECT_TYPE_RELEASE) {
                $release= $releaseRepository->findOneBy(array('project' => $project));
                $zhUserAllocationService->createManagerRelease($release, $zhUsers);
            } else if ($externalType == self::PROJECT_TYPE_SUB_RELEASE) {
                $subRelease = InstanceProvider::getRepository()->getSubrelease()->findOneBy(array('project' => $project));
                $zhUserAllocationService->createManagerSubRelease($subRelease, $zhUsers);
                $release = $releaseRepository->findOneBy(array('project' => $project));
                $zhUserAllocationService->createManagerRelease($release, $zhUsers);
            } else if ($externalType == self::PROJECT_TYPE_ERROR) {
                $subRelease = InstanceProvider::getRepository()->getSubrelease()->findOneBy(array('project' => $project));
                $zhUserAllocationService->createManagerSubRelease($subRelease, $zhUsers);
            }
        }
    }

    private function validateRelationshipModelRestricted(Entities\Project $project)
    {
        $modelProjectPrime = InstanceProvider::getService()->getProjectRelationship()->getOneProjectRelationship($project,
            Entities\ProjectRelationship::TYPE_MODEL_GENERATION);
        return $modelProjectPrime
        && $modelProjectPrime->getProjectType()->getExternaltype()
        == Entities\ProjectType::TYPE_PROJECT_CONTROL_MODEL_RESTRICTED;

    }

    /**
     * @param Entities\Project $modelProject
     * @param $controlsNotCreated
     * @param $taskInDeveloping
     * @param $taskForMerge
     * @return array
     */
    private function setsModelProject(Entities\Project $modelProject, $controlsNotCreated, $taskInDeveloping, $taskForMerge)
    {
        $modelProjectPrime = $this->getModelProjectParent($modelProject);
        if ($modelProjectPrime
            && $modelProjectPrime->getProjectType()
            && $modelProjectPrime->getProjectType()->getExternaltype() == Entities\ProjectType::TYPE_PROJECT_CONTROL_MODEL_RESTRICTED
            && $modelProject->getProjectStatus()
            && $modelProject->getProjectStatus()->getControl() != Entities\TaskStatus::TASK_STATUS_CONTROL_PLANNING
        ) {
            $projects = InstanceProvider::getService()->getProjectRelationship()->getProjects($modelProject,
                Entities\ProjectRelationship::TYPE_MODEL);
            foreach ($projects as $project) {
                if($project->getProjectStatus() &&  (!in_array($project->getProjectStatus()->getControl(), self::arrayStatusNotModify() ))){
                    $project->setProjectStatus($modelProject->getProjectStatus());
                    $this->validateIsPossibleChangeForInHomologation($project, $controlsNotCreated, $taskInDeveloping, $taskForMerge);
                    $this->entityManager->persist($project);
                }
            }
            if (empty($controlsNotCreated) && empty($taskInDeveloping) && empty($taskForMerge)) {
                $this->entityManager->flush();
                foreach ($projects as $project) {
                    $this->setParentStatusProject($project);
                    $this->setsDevOps($project);
                    $this->setChildrenStatusProject($project, true);
                    $this->setsModelProject($project, $controlsNotCreated, $taskInDeveloping, $taskForMerge);
                }
            }
        }
        return array($controlsNotCreated, $taskInDeveloping, $taskForMerge);
    }

    /**
     * @param Entities\Project $modelProject
     * @param $controlsNotCreated
     * @param $taskInDeveloping
     * @param $taskForMerge
     * @throws \Exception
     */
    private function validateHomolagation(Entities\Project $modelProject, $controlsNotCreated, $taskInDeveloping, $taskForMerge)
    {
        if (!(empty($controlsNotCreated) && empty($taskInDeveloping) && empty($taskForMerge))) {
            $labelStatus = $modelProject->getProjectStatus()->getLabel();
            $messageError = 'Para utilizar o status de projetos "' . $labelStatus . '" é necessário: <BR><BR>';
            if (!empty($controlsNotCreated)) {
                $stringCategory = $stringCategory = implode('","', $controlsNotCreated);
                $messageError .= '<p><u>Criar a(s) categoria(s) de tarefas com os controle(s) "' . $stringCategory . '".</u></p><BR>';
            }

            if (!empty($taskInDeveloping)) {
                $messageTask = '<p><u>O(s) projeto(s) a seguir não pode(m) ter tarefas em andamento: </u></p> <BR>';
                foreach ($taskInDeveloping as $key => $tasksName) {
                    $taskName = implode('<BR>Tarefa: ', $tasksName);
                    $messageTask .= "Projeto: $key <BR>Tarefa: $taskName <BR>";
                }

                $messageError .= $messageTask . '<BR>';
            }

            if (!empty($taskForMerge)) {
                $messageTask = '<p><u>O(s) projeto(s) a seguir não pode(m) ter Merge Requests pendentes: </u></p> <BR>';
                foreach ($taskForMerge as $key => $tasksName) {
                    $taskName = implode('<BR>Tarefa: ', $tasksName);
                    $messageTask .= "Projeto: $key <BR>Tarefa: $taskName <BR>";
                }

                $messageError .= $messageTask . '<BR>';
            }

            throw new \Exception($messageError);
        }
    }


    /**
     * @param Entities\Project $modelProject
     * @throws \Exception
     */
    private function validateConcluded(Entities\Project $modelProject)
    {
        if ($modelProject->getProjectStatus()->getControl() == Entities\TaskStatus::TASK_STATUS_CONTROL_CONCLUDED) {

            $projects = InstanceProvider::getRepository()->getProject()->getAllProjectRelationshipRestricted($modelProject->getId());
            $validate = true;
            /** @var Entities\Project[] $allProjectsNotConcluded */
            $allProjectsNotConcluded = [];
            foreach ($projects as $project) {
                if($project->getProjectStatus() && $project->getProjectStatus()->getControl() != Entities\TaskStatus::TASK_STATUS_CONTROL_CONCLUDED && $project->getProjectStatus()->getControl() != Entities\TaskStatus::TASK_STATUS_CONTROL_CANCELED){
                    $validate = false;
                    $allProjectsNotConcluded[] = $project;
                }
            }

            if (!$validate) {
                $messageError = Messages::ERR_CONCLUDED_MODEL_RESTRICTED;
                foreach ($allProjectsNotConcluded as $project) {
                    $projectName = $project->getId().' - '.$project->getName();
                    $messageError .=  "<p>$projectName</p>";
                }
                throw new \Exception($messageError);
            }
        }
    }

    /**
     * @param Entities\Project $project
     * @throws \Exception
     */
    private function deleteProjectModel($project)
    {
        $projectModel = $this->getModelProjectParent($project);
        if ($projectModel
            && $projectModel->getProjectType()
            && $projectModel->getProjectType()->getExternaltype() == Entities\ProjectType::TYPE_PROJECT_CONTROL_MODEL_RESTRICTED
        ) {
            $projectIdChildren = DevOpsUtilities::getAllProjectLessIdInArray($project->getId());
            $projectChildrenGenerations = InstanceProvider::getService()->getProjectRelationship()->getProjects($projectIdChildren,
                Entities\ProjectRelationship::TYPE_MODEL);
            $tasks = InstanceProvider::getRepository()->getTask()->findBy(array('project' => $projectChildrenGenerations));
            if (!empty($tasks)) {
                throw new \Exception(Messages::ERR_REMOVE_PROJECT_MODEL_WITH_TASK);
            }
            foreach ($projectChildrenGenerations as $projectChildrenGeneration) {
                $this->deleteDevOpsAssociation($project);
                $projectChildrenGeneration->setStatus(LogicalDeleted::STATUS_LOGICAL_DELETED);
                $this->entityManager->persist($projectChildrenGeneration);
                $this->deleteProjectModel($projectChildrenGeneration);
            }
        }
    }

    /**
     * @param Entities\Project $targetProject
     * @param Entities\Task $task
     */
    private function updateRepositoriesTaskChangedFiles(Entities\Project $targetProject, Entities\Task $task) {
        $repositoryRepository = InstanceProvider::getRepository()->getRepository();
        /** @var Entities\Repository[] $targetRepositories */
        $targetRepositories = $repositoryRepository->findBy(array('project' => $targetProject));
        /** @var Entities\ChangedFiles[] $changedFiles */
        $changedFiles = InstanceProvider::getRepository()->getChangedFiles()->findBy(array('task' => $task));

        foreach ($changedFiles as $changedFile) {
            $isUpdated = false;
            /** @var Entities\Repository $taskRepository */
            $taskRepository = $repositoryRepository->find($changedFile->getRepository()->getId());
            foreach($targetRepositories as $targetRepository) {
                if($taskRepository->getUrl() == $targetRepository->getUrl()) {
                    $changedFile->setRepository($targetRepository);
                    $isUpdated = true;
                    break;
                }
            }
            if($isUpdated) {
                $this->entityManager->persist($changedFile);
            }
        }
    }

    /**
     * @param Entities\Project $project
     * @param $taskInDeveloping
     * @param $taskForMerge
     * @param Entities\Task $task
     * @param $arrayClassificationNotValidation
     * @return array
     */
    private function validateTaskChangeInHomologation(Entities\Project $project, &$taskInDeveloping, &$taskForMerge,
                                                      Entities\Task $task, $arrayClassificationNotValidation)
    {
        if ($task->getTaskStatus() &&
            $task->getTaskStatus()->getControl() != Entities\TaskStatus::TASK_STATUS_CONTROL_CONCLUDED &&
            $task->getClassification() &&
            (!in_array($task->getClassification()->getControl(), $arrayClassificationNotValidation) || !$task->getClassification()->getControl())&&
            !$this->validateMergeRequestChangeInHomologation($task)
        ) {
            if ($task->getTaskStatus()->getControl() == Entities\TaskStatus::TASK_STATUS_CONTROL_IN_DEVELOPING) {
                $taskInDeveloping[$project->getId() . ' - ' . $project->getName()][] = $task->getId() . ' - ' . $task->getName();
            } else if ($task->getClassification() &&
                $task->getClassification()->getControl() == Entities\Classification::CONTROL_MERGE_REQUEST
            ) {
                $taskForMerge[$project->getId() . ' - ' . $project->getName()][] = $task->getId() . ' - ' . $task->getName();
            }
        }
        return array($taskInDeveloping, $taskForMerge);
    }


    private function validateMergeRequestChangeInHomologation(Entities\Task $task)
    {
        $validation = false;
        $mergeTaskRelationship = $this->findTaskRelationshipByTask($task,Entities\TaskRelationship::MERGE_REQUEST_TYPE);
        if($mergeTaskRelationship){
            $arrayTypes = array(Entities\TaskRelationship::ERROR_TYPE,Entities\TaskRelationship::CONFLICT_DEVOPS_TYPE);
            $taskRelationship = $this->findTaskRelationshipByTask($mergeTaskRelationship->getTask(),$arrayTypes);
            if($taskRelationship && $taskRelationship->getType() == Entities\TaskRelationship::ERROR_TYPE){
                $testTaskRelationship = $this->findTaskRelationshipByTask($taskRelationship->getTask(),Entities\TaskRelationship::TEST_TYPE);
                if($testTaskRelationship){
                    $validation = true;
                }
            }else if($taskRelationship && $taskRelationship->getType() == Entities\TaskRelationship::CONFLICT_DEVOPS_TYPE){
                $validation = true;
            }
        }
        return $validation;
    }

    /**
     * @param Entities\Task $task
     * @param $type
     * @return null|Entities\TaskRelationship
     */
    private function findTaskRelationshipByTask(Entities\Task $task,$type)
    {
        $taskRelationship = InstanceProvider::getRepository()->getTaskRelationship()->findOneBy(
            array('taskRelationship' => $task, 'type' => $type)
        );
        return $taskRelationship;
    }

    public function validateConcludedTaskToMove($tasks, $projectId, $targetProjectId)
    {
        try {
            $concludedTasks = [];
            foreach ($tasks as $task) {
                if(($task['PERCENT_COMPLETED'] == 100 ||
                    $task['TASK_STATUS'] == Utilities::getTaskStatusIsConcluded()->getId()
                    ) && $projectId != $targetProjectId) {
                    array_push($concludedTasks, $task['NAME']);
                }
            }

            $response = [];
            if(empty($concludedTasks)) {
                $taskParentIds = [];
                foreach ($tasks as $task) {
                    $taskParentIds[] = $task['ID'];
                }
                $taskParentIdsConcat = join(',', $taskParentIds);
                $filter = "AND P.PERCENT_COMPLETED = '100'";
                $concludedChildTasks = DevOpsUtilities::getAllTaskChildrenIdInArray($taskParentIdsConcat, $filter);
                if(!empty($concludedChildTasks)){
                    $response = array('CONCLUDED_TASK_CHILD' => $concludedChildTasks);
                }

            } else {
                $response = array('ERROR' => $concludedTasks);
            }
            return  [$response];
        } catch (\Exception $error) {
            throw $error;
        }
    }

    public function findAllProjects($filters, $page, $pageSize)
    {
        try {
            $limit = $page * $pageSize;
            $offset = ($page - 1) * $pageSize;
            $filter = DevOpsUtilities::queryFilterTypeProjectMeeting('PET', true);
            $filters = DevOpsUtilities::createQueryFiltered($filters, 'PRO');
            return InstanceProvider::getRepository()->getProject()->findAllProjects($filter, $limit, $offset, $filters);
        }catch(\Exception $e){
            throw $e;
        }
    }

    public function hasErrorTypeChildrenProjects($conditions)
    {
        $hasTypeProjectError = false;
        $projectId = DevOpsUtilities::getValueInFilterDataSourceByName($conditions, 'PROJECT_ID');
        $childrenProjects = InstanceProvider::getRepository()->getProject()->findBy(array('parent' =>  $projectId));
        if(!empty($childrenProjects)){
            $hasTypeProjectError = $this->findErrorTypeChildrenProjects($childrenProjects);
        }
        return $hasTypeProjectError;
    }

    public function findErrorTypeChildrenProjects($childrenProjects)
    {
        foreach ($childrenProjects as $project) {
            $projectType = InstanceProvider::getRepository()->getProjectType()->find($project->getProjectType());
            if($projectType->getExternaltype() == 'E'){
                return true;
            } else {
                $childrenProjects = InstanceProvider::getRepository()->getProject()->findBy(array('parent' =>  $project->getId()));
                if(!empty($childrenProjects)){
                    return $this->findErrorTypeChildrenProjects($childrenProjects);
                }
            }
        }
        return false;
    }

    public function projectHasTask($projectId, $conditions = array())
    {
        $organizationId = $this->environment->getCurrentOrganizationId();
        $filter = DevOpsUtilities::createQueryFiltered($conditions, 'T');
        $task = InstanceProvider::getRepository()->getTask()->findTaskByProject($projectId, $filter, $organizationId);
        if(!empty($task)){
            return true;
        }
        return false;
    }

    public function findTaskResponsesRelatedSameProject($projectId, $conditions = array())
    {
        return InstanceProvider::getRepository()->getTask()->findTaskResponsesHomologation($projectId);

    }

    public function findTaskResponsesRelatedSameIssue($projectId, $conditions = array())
    {
        $organizationId = $this->environment->getCurrentOrganizationId();
        return InstanceProvider::getRepository()->getTask()->findTaskResponsesRelatedSameIssue($projectId, $organizationId);

    }

    public function hasProjectWithLtsSubReleaseFinalized($projectId)
    {
        $project = InstanceProvider::getRepository()->getProject()->find($projectId);
        $subreleases = InstanceProvider::getRepository()->getSubrelease()->findBy(array('project' => $project->getId()));
        $isSubrelease = false;
        foreach ($subreleases as $s) {
            if($s->getProductVersion()){

                $productVersion = InstanceProvider::getRepository()->getProductVersion()->find($s->getProductVersion()->getId());
                if($productVersion->getIsLts() == 'Y' && $s->getStatus() == ConfigurationmanagerEntities\Subrelease::HOMOLOGATED_STATUS || $s->getStatus() == ConfigurationmanagerEntities\Subrelease::TESTED_STATUS){
                    $isSubrelease = true;
                }
            }
        }
        return $isSubrelease;
    }

    public function hasErrorTypeChildrenProjectsAndHasProjectWithLtsSubReleaseFinalized($projectId)
    {
        $data = false;
        $filter = "AND PT.EXTERNALTYPE = 'E'";
        $projects = InstanceProvider::getRepository()->getProject()->getProjectChildrenNotConcludedOrCanceled($projectId, $filter);
        if(!empty($projects)){
            foreach ($projects as $project) {
                $data = $this->hasProjectWithLtsSubReleaseFinalized($project);
                if($data) break;
            }

        }
        return $data;
    }

    /**
     * @param $projectId
     * @param $parentProjectId
     * @return array
     */
    public function validateParentProjectChoiceByUser($projectId, $parentProjectId)
    {
        $data = array();
        if($parentProjectId){
            $data['IS_MANAGER'] = $this->validateParentProjectByManager($parentProjectId);
            $data['IS_CHILDREN'] = $this->validateParentProjectChoiceByProject($projectId, $parentProjectId);
        } else {
            $data['IS_MANAGER'] = true;
            $data['IS_CHILDREN'] = false;
        }
        return $data;
    }

    private function validateParentProjectChoiceByProject($projectId, $parentProjectId)
    {
        $isChildren = InstanceProvider::getRepository()->getProject()->validateParentProjectChoiceByProject($projectId, $parentProjectId);
        return empty($isChildren) ? false : true;
    }

    private function validateParentProjectByManager($parentProjectId)
    {
        $projectRepository =  InstanceProvider::getRepository()->getProject();

        $resourceId = $this->environment->getCurrentResourceId();
        $resourceStructureManagerIds = InstanceProvider::getRepository()->getResource()->findManagerByResourceId($resourceId);
        $projects = $projectRepository->findProjectsParentByProjectId($parentProjectId);
        $managerIds = $this->createArrayManagerIds($resourceStructureManagerIds, $resourceId);

        $isManager = $this->isManagersInProjects($projects, $managerIds);

        if($isManager){
            return true;
        } else {
            $structureIds = $this->createArrayStructureIds($projects);
            $structureIdsString = implode(', ', $structureIds);
            $structuresManagers = $projectRepository->findStructureParentByStructureIds($structureIdsString);

            $structureIds = array();
            foreach ($structuresManagers as $structure){
                array_push($structureIds, $structure['RESOURCE_T']);
            }
            $isManager = array_intersect($managerIds, $structureIds);
            return empty($isManager) ? false : true;
        }
    }

    private function createArrayManagerIds($managersId, $resourceId)
    {
        $ids = array();
        foreach ($managersId as $id){
            array_push($ids, $id['MANAGER']);
        }
        array_push($ids, $resourceId);
        return $ids;
    }

    private function createArrayStructureIds($projects)
    {
        $structure = array();
        foreach ($projects as $project){
            array_push($structure, $project['STRUCTURE']);
        }
        return array_unique($structure);
    }

    private function isManagersInProjects($projects, $managersId)
    {
        $managersString = '';
        foreach ($projects as $project){
            $managersString .= $project['MANAGER'] . ',';
        }
        $managers = explode(',', $managersString);
        $result = array_intersect($managersId, $managers);
        return (empty($result)) ? false : true;
    }


}
